# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class DocumentProperty(JsonProperty):
    SPDX_ID = auto()
    ANNOTATIONS = auto()
    COMMENT = auto()
    CREATION_INFO = auto()
    DATA_LICENSE = auto()
    EXTERNAL_DOCUMENT_REFS = auto()
    HAS_EXTRACTED_LICENSING_INFOS = auto()
    NAME = auto()
    SPDX_VERSION = auto()
    DOCUMENT_NAMESPACE = auto()
    PACKAGES = auto()
    FILES = auto()
    SNIPPETS = auto()
    RELATIONSHIPS = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict, Optional

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import json_str_to_enum_name
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)


class ChecksumParser:
    logger: Logger

    def __init__(self):
        self.logger = Logger()

    @staticmethod
    def parse_checksum(checksum_dict: Dict) -> Checksum:
        logger = Logger()
        algorithm: str = json_str_to_enum_name(checksum_dict.get("algorithm", ""))
        try:
            checksum_algorithm = ChecksumAlgorithm[algorithm]
        except KeyError:
            logger.append(f"Invalid ChecksumAlgorithm: {algorithm}")
            checksum_algorithm = None
        checksum_value: Optional[str] = checksum_dict.get("checksumValue")
        raise_parsing_error_if_logger_has_messages(logger, "Checksum")
        checksum = construct_or_raise_parsing_error(Checksum, dict(algorithm=checksum_algorithm, value=checksum_value))
        return checksum



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import IntegrityMethod
from spdx_tools.spdx3.writer.console.console import write_value


def write_integrity_method(integrity_method: IntegrityMethod, text_output: TextIO, indent: bool = True):
    write_value("comment", integrity_method.comment, text_output, indent)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class ChecksumProperty(JsonProperty):
    ALGORITHM = auto()
    CHECKSUM_VALUE = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.license_expression_parser import LicenseExpressionParser


@pytest.mark.parametrize(
    "license_expression_str, expected_license",
    [
        ("First License", spdx_licensing.parse("First License")),
        ("Second License", spdx_licensing.parse("Second License")),
        ("Apache-1.1", spdx_licensing.parse("Apache-1.1")),
        ("Zlib", spdx_licensing.parse("zlib")),
        ("NOASSERTION", SpdxNoAssertion()),
        ("NONE", SpdxNone()),
    ],
)
def test_parse_license_expression(license_expression_str, expected_license):
    license_expression_parser = LicenseExpressionParser()
    license_expression = license_expression_parser.parse_license_expression(license_expression_str)

    assert license_expression == expected_license


@pytest.mark.parametrize(
    "invalid_license_expression,expected_message",
    [
        (56, ["Error parsing LicenseExpression: \"56\": expression must be a string and not: <class 'int'>"]),
        (
            "LGPL-2.1, GPL-2.0, GPL-3.0",
            [
                # the error message we receive from the license_expression library somehow cuts off the last license
                "Error parsing LicenseExpression: \"LGPL-2.1, GPL-2.0, GPL-3.0\": Invalid license key: the valid characters are: letters and numbers, underscore, dot, colon or hyphen signs and spaces: 'LGPL-2.1, GPL-2.0,'"  # noqa: E501
            ],
        ),
        ("Apache License (2.0)", ['Error parsing LicenseExpression: "Apache License (2.0)"']),
    ],
)
def test_parse_invalid_license_expression(invalid_license_expression, expected_message):
    license_expression_parser = LicenseExpressionParser()

    with pytest.raises(SPDXParsingError) as err:
        license_expression_parser.parse_license_expression(invalid_license_expression)

    TestCase().assertCountEqual(err.value.get_messages(), expected_message)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto

from beartype.typing import Dict, List, Optional, Tuple, Union
from license_expression import LicenseExpression

from spdx_tools.spdx.model import Snippet, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import (
    parse_field_or_log_error,
    parse_field_or_no_assertion_or_none,
)
from spdx_tools.spdx.parser.jsonlikedict.license_expression_parser import LicenseExpressionParser
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import construct_or_raise_parsing_error


class RangeType(Enum):
    BYTE = auto()
    LINE = auto()


class SnippetParser:
    logger: Logger
    license_expression_parser = LicenseExpressionParser

    def __init__(self):
        self.logger = Logger()
        self.license_expression_parser = LicenseExpressionParser()

    def parse_snippet(self, snippet_dict: Dict) -> Snippet:
        logger = Logger()
        spdx_id: Optional[str] = snippet_dict.get("SPDXID")
        file_spdx_id: Optional[str] = snippet_dict.get("snippetFromFile")
        name: Optional[str] = snippet_dict.get("name")

        ranges: Dict = parse_field_or_log_error(logger, snippet_dict.get("ranges", []), self.parse_ranges, default={})
        byte_range: Optional[Tuple[Union[int, str], Union[int, str]]] = ranges.get(RangeType.BYTE)
        line_range: Optional[Tuple[Union[int, str], Union[int, str]]] = ranges.get(RangeType.LINE)
        byte_range = self.convert_range_from_str(byte_range)
        line_range = self.convert_range_from_str(line_range)

        attribution_texts: List[str] = snippet_dict.get("attributionTexts", [])
        comment: Optional[str] = snippet_dict.get("comment")
        copyright_text: Optional[Union[str, SpdxNoAssertion, SpdxNone]] = parse_field_or_no_assertion_or_none(
            snippet_dict.get("copyrightText")
        )
        license_comment: Optional[str] = snippet_dict.get("licenseComments")
        license_concluded: Optional[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]] = parse_field_or_log_error(
            logger, snippet_dict.get("licenseConcluded"), self.license_expression_parser.parse_license_expression
        )

        license_info: List[Union[LicenseExpression], SpdxNoAssertion, SpdxNone] = parse_field_or_log_error(
            logger,
            snippet_dict.get("licenseInfoInSnippets"),
            self.license_expression_parser.parse_license_expression,
            field_is_list=True,
        )
        if logger.has_messages():
            raise SPDXParsingError([f"Error while parsing snippet: {logger.get_messages()}"])

        snippet = construct_or_raise_parsing_error(
            Snippet,
            dict(
                spdx_id=spdx_id,
                name=name,
                byte_range=byte_range,
                file_spdx_id=file_spdx_id,
                line_range=line_range,
                attribution_texts=attribution_texts,
                comment=comment,
                copyright_text=copyright_text,
                license_comment=license_comment,
                license_concluded=license_concluded,
                license_info_in_snippet=license_info,
            ),
        )

        return snippet

    def parse_ranges(self, ranges_from_snippet: List[Dict]) -> Dict:
        logger = Logger()
        ranges = {}
        for range_dict in ranges_from_snippet:
            try:
                range_type: RangeType = self.validate_range_and_get_type(range_dict)
                start_end_tuple: Tuple[int, int] = SnippetParser.get_start_end_tuple(range_dict, range_type)
                ranges[range_type] = start_end_tuple
            except ValueError as error:
                logger.append(error.args[0])
        if logger.has_messages():
            raise SPDXParsingError([f"Error while parsing snippet ranges: {logger.get_messages()}"])
        return ranges

    @staticmethod
    def get_start_end_tuple(range_dict: Dict, range_type: RangeType) -> Tuple[int, int]:
        end_pointer: Dict = range_dict["endPointer"]
        start_pointer: Dict = range_dict["startPointer"]
        if range_type == RangeType.BYTE:
            start: int = start_pointer["offset"]
            end: int = end_pointer["offset"]
        else:
            start: int = start_pointer["lineNumber"]
            end: int = end_pointer["lineNumber"]
        return start, end

    def validate_range_and_get_type(self, range_dict: Dict) -> RangeType:
        if "startPointer" not in range_dict:
            raise ValueError("Startpointer missing in snippet ranges.")
        if "endPointer" not in range_dict:
            raise ValueError("Endpointer missing in snippet ranges.")
        start_pointer_type: RangeType = self.validate_pointer_and_get_type(range_dict["startPointer"])
        end_pointer_type: RangeType = self.validate_pointer_and_get_type(range_dict["endPointer"])
        if start_pointer_type != end_pointer_type:
            raise ValueError("Type of startpointer is not the same as type of endpointer.")
        return start_pointer_type

    @staticmethod
    def validate_pointer_and_get_type(pointer: Dict) -> RangeType:
        if "offset" in pointer and "lineNumber" in pointer:
            raise ValueError('Couldn\'t determine type of pointer: "offset" and "lineNumber" provided as key.')
        if "offset" not in pointer and "lineNumber" not in pointer:
            raise ValueError('Couldn\'t determine type of pointer: neither "offset" nor "lineNumber" provided as key.')
        return RangeType.BYTE if "offset" in pointer else RangeType.LINE

    @staticmethod
    def convert_range_from_str(
        _range: Tuple[Union[int, str], Union[int, str]]
    ) -> Tuple[Union[int, str], Union[int, str]]:
        # XML does not support integers, so we have to convert from string (if possible)
        if not _range:
            return _range

        if isinstance(_range[0], str) and _range[0].isdigit():
            _range = int(_range[0]), _range[1]
        if isinstance(_range[1], str) and _range[1].isdigit():
            _range = _range[0], int(_range[1])
        return _range



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.model import ChecksumAlgorithm
from spdx_tools.spdx.writer.tagvalue.checksum_writer import write_checksum_to_tag_value
from tests.spdx.fixtures import checksum_fixture


@pytest.mark.parametrize(
    "checksum, expected_string",
    [
        (checksum_fixture(), "SHA1: 71c4025dd9897b364f3ebbb42c484ff43d00791c"),
        (checksum_fixture(algorithm=ChecksumAlgorithm.SHA3_256, value="fdsef"), "SHA3-256: fdsef"),
        (checksum_fixture(algorithm=ChecksumAlgorithm.SHA3_384, value="fdsef"), "SHA3-384: fdsef"),
        (checksum_fixture(algorithm=ChecksumAlgorithm.SHA3_512, value="fdsef"), "SHA3-512: fdsef"),
        (checksum_fixture(algorithm=ChecksumAlgorithm.BLAKE2B_256, value="fdsef"), "BLAKE2b-256: fdsef"),
        (checksum_fixture(algorithm=ChecksumAlgorithm.BLAKE2B_384, value="fdsef"), "BLAKE2b-384: fdsef"),
        (checksum_fixture(algorithm=ChecksumAlgorithm.BLAKE2B_512, value="fdsef"), "BLAKE2b-512: fdsef"),
    ],
)
def test_checksum_writer(checksum, expected_string):
    checksum_string = write_checksum_to_tag_value(checksum)

    assert checksum_string == expected_string



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import json
import os

import pytest

from spdx_tools.spdx.writer.json.json_writer import write_document_to_file
from tests.spdx.fixtures import document_fixture


@pytest.fixture
def temporary_file_path() -> str:
    temporary_file_path = "temp_test_json_writer_output.json"
    yield temporary_file_path
    os.remove(temporary_file_path)


def test_write_json(temporary_file_path: str):
    document = document_fixture()
    write_document_to_file(document, temporary_file_path, validate=True)

    with open(temporary_file_path, encoding="utf-8") as written_file:
        written_json = json.load(written_file)

    with open(
        os.path.join(os.path.dirname(__file__), "expected_results", "expected.json"), encoding="utf-8"
    ) as expected_file:
        expected_json = json.load(expected_file)

    assert written_json == expected_json


def test_document_is_validated():
    document = document_fixture()
    document.creation_info.spdx_id = "InvalidId"

    with pytest.raises(ValueError) as error:
        write_document_to_file(document, "dummy_path")
    assert "Document is not valid" in error.value.args[0]


def test_document_validation_can_be_overridden(temporary_file_path: str):
    document = document_fixture()
    document.creation_info.spdx_id = "InvalidId"

    write_document_to_file(document, temporary_file_path, validate=False)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import IO, Dict
from rdflib import DOAP, Graph
from rdflib.compare import to_isomorphic

from spdx_tools.spdx.model import Document
from spdx_tools.spdx.rdfschema.namespace import POINTER_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.annotation_writer import add_annotation_to_graph
from spdx_tools.spdx.writer.rdf.creation_info_writer import add_creation_info_to_graph
from spdx_tools.spdx.writer.rdf.extracted_licensing_info_writer import add_extracted_licensing_info_to_graph
from spdx_tools.spdx.writer.rdf.file_writer import add_file_to_graph
from spdx_tools.spdx.writer.rdf.package_writer import add_package_to_graph
from spdx_tools.spdx.writer.rdf.relationship_writer import add_relationship_to_graph
from spdx_tools.spdx.writer.rdf.snippet_writer import add_snippet_to_graph
from spdx_tools.spdx.writer.write_utils import validate_and_deduplicate


def write_document_to_stream(
    document: Document, stream: IO[bytes], validate: bool = True, drop_duplicates: bool = True
):
    document = validate_and_deduplicate(document, validate, drop_duplicates)
    graph = Graph()
    doc_namespace = document.creation_info.document_namespace
    external_doc_ref_to_namespace: Dict[str, str] = {
        external_doc_ref.document_ref_id: external_doc_ref.document_uri
        for external_doc_ref in document.creation_info.external_document_refs
    }
    doc_node = add_creation_info_to_graph(document.creation_info, graph)
    for annotation in document.annotations:
        add_annotation_to_graph(annotation, graph, doc_namespace, external_doc_ref_to_namespace)

    for file in document.files:
        add_file_to_graph(file, graph, doc_namespace, external_doc_ref_to_namespace)

    for package in document.packages:
        add_package_to_graph(package, graph, doc_namespace, external_doc_ref_to_namespace)

    for relationship in document.relationships:
        add_relationship_to_graph(relationship, graph, doc_namespace, external_doc_ref_to_namespace)

    for snippet in document.snippets:
        add_snippet_to_graph(snippet, graph, doc_namespace, external_doc_ref_to_namespace)

    for extracted_licensing_info in document.extracted_licensing_info:
        add_extracted_licensing_info_to_graph(extracted_licensing_info, graph, doc_node, doc_namespace)

    graph = to_isomorphic(graph)
    graph.bind("spdx", SPDX_NAMESPACE)
    graph.bind("doap", DOAP)
    graph.bind("ptr", POINTER_NAMESPACE)
    graph.serialize(stream, "pretty-xml", encoding="UTF-8", max_depth=100)


def write_document_to_file(document: Document, file_name: str, validate: bool = True, drop_duplicates: bool = True):
    with open(file_name, "wb") as out:
        write_document_to_stream(document, out, validate, drop_duplicates)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock, call, mock_open, patch

import pytest

from spdx_tools.spdx.model import ActorType, RelationshipType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import scan_relationships, write_actor
from tests.spdx.fixtures import actor_fixture, file_fixture, package_fixture, relationship_fixture


def test_scan_relationships():
    first_package_spdx_id = "SPDXRef-Package1"
    second_package_spdx_id = "SPDXRef-Package2"
    packages = [package_fixture(spdx_id=first_package_spdx_id), package_fixture(spdx_id=second_package_spdx_id)]
    file_spdx_id = "SPDXRef-File"
    files = [file_fixture(spdx_id=file_spdx_id)]
    no_assertion_relationship = relationship_fixture(
        spdx_element_id=second_package_spdx_id,
        relationship_type=RelationshipType.CONTAINS,
        related_spdx_element_id=SpdxNoAssertion(),
        comment=None,
    )
    none_relationship = relationship_fixture(
        spdx_element_id=second_package_spdx_id,
        relationship_type=RelationshipType.CONTAINS,
        related_spdx_element_id=SpdxNone(),
        comment=None,
    )
    relationships = [
        relationship_fixture(
            spdx_element_id=first_package_spdx_id,
            relationship_type=RelationshipType.CONTAINS,
            related_spdx_element_id=file_spdx_id,
            comment=None,
        ),
        relationship_fixture(
            spdx_element_id=second_package_spdx_id,
            relationship_type=RelationshipType.CONTAINS,
            related_spdx_element_id=file_spdx_id,
            comment=None,
        ),
        no_assertion_relationship,
        none_relationship,
    ]

    relationships_to_write, contained_files_by_package_id = scan_relationships(relationships, packages, files)

    assert relationships_to_write == [no_assertion_relationship, none_relationship]
    assert contained_files_by_package_id == {first_package_spdx_id: files, second_package_spdx_id: files}


@pytest.mark.parametrize(
    "element_to_write, expected_calls",
    [
        (actor_fixture(), [call("ActorTest: Person: actorName (some@mail.com)\n")]),
        (
            actor_fixture(actor_type=ActorType.ORGANIZATION, name="organizationName"),
            [call("ActorTest: Organization: organizationName (some@mail.com)\n")],
        ),
        (actor_fixture(actor_type=ActorType.TOOL, name="toolName", email=None), [call("ActorTest: Tool: toolName\n")]),
        (SpdxNoAssertion(), [call("ActorTest: NOASSERTION\n")]),
    ],
)
def test_write_actor(element_to_write, expected_calls):
    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_actor("ActorTest", element_to_write, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(expected_calls)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

import pytest

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.annotation_properties import AnnotationProperty
from spdx_tools.spdx.model import Actor, ActorType, Annotation, AnnotationType


@pytest.fixture
def converter() -> AnnotationConverter:
    return AnnotationConverter()


@pytest.mark.parametrize(
    "annotation_property,expected",
    [
        (AnnotationProperty.ANNOTATION_DATE, "annotationDate"),
        (AnnotationProperty.ANNOTATION_TYPE, "annotationType"),
        (AnnotationProperty.ANNOTATOR, "annotator"),
        (AnnotationProperty.COMMENT, "comment"),
    ],
)
def test_json_property_names(converter: AnnotationConverter, annotation_property: AnnotationProperty, expected: str):
    assert converter.json_property_name(annotation_property) == expected


def test_json_type(converter: AnnotationConverter):
    assert converter.get_json_type() == AnnotationProperty


def test_data_model_type(converter: AnnotationConverter):
    assert converter.get_data_model_type() == Annotation


def test_successful_conversion(converter: AnnotationConverter):
    date = datetime(2022, 12, 1)
    annotator = Actor(ActorType.PERSON, "actorName")
    annotation = Annotation("spdxId", AnnotationType.REVIEW, annotator, date, "comment")

    converted_dict = converter.convert(annotation)

    assert converted_dict == {
        converter.json_property_name(AnnotationProperty.ANNOTATION_DATE): datetime_to_iso_string(date),
        converter.json_property_name(AnnotationProperty.ANNOTATION_TYPE): "REVIEW",
        converter.json_property_name(AnnotationProperty.ANNOTATOR): annotator.to_serialized_string(),
        converter.json_property_name(AnnotationProperty.COMMENT): "comment",
    }



#  SPDX-FileCopyrightText: 2023 spdx contributors
#
#  SPDX-License-Identifier: Apache-2.0
import logging
from os import path

from spdx_tools.spdx.model.document import Document
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.parse_anything import parse_file

# This example demonstrates how to parse an existing spdx file.

# Provide a path to the input file
input_path = path.join(path.dirname(__file__), "..", "tests", "spdx", "data", "SPDXLite.spdx")
try:
    # Try to parse the input file. If successful, returns a Document, otherwise raises an SPDXParsingError
    document: Document = parse_file(input_path)
except SPDXParsingError:
    logging.exception("Failed to parse spdx file")

# We can now access attributes from the parsed document
print(f"Parsed document name: {document.creation_info.name}")
creators_as_str = ", ".join([creator.to_serialized_string() for creator in document.creation_info.creators])
print(f"Created on {document.creation_info.created} by {creators_as_str}")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import Checksum


@dataclass_with_properties
class ExternalDocumentRef:
    document_ref_id: str  # of the form "DocumentRef-[idstring]"
    document_uri: str
    checksum: Checksum

    def __init__(self, document_ref_id: str, document_uri: str, checksum: Checksum):
        check_types_and_set_values(self, locals())



#  SPDX-FileCopyrightText: 2023 spdx contributors
#
#  SPDX-License-Identifier: Apache-2.0
import logging
from datetime import datetime
from typing import List

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Checksum,
    ChecksumAlgorithm,
    CreationInfo,
    Document,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    File,
    FileType,
    Package,
    PackagePurpose,
    PackageVerificationCode,
    Relationship,
    RelationshipType,
)
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.validation.validation_message import ValidationMessage
from spdx_tools.spdx.writer.write_anything import write_file

# This example shows how to use the spdx-tools to create an SPDX document from scratch,
# validate it and write it to a file.

# First up, we need general information about the creation of the document, summarised by the CreationInfo class.
creation_info = CreationInfo(
    spdx_version="SPDX-2.3",
    spdx_id="SPDXRef-DOCUMENT",
    name="document name",
    data_license="CC0-1.0",
    document_namespace="https://some.namespace",
    creators=[Actor(ActorType.PERSON, "Jane Doe", "jane.doe@example.com")],
    created=datetime(2022, 1, 1),
)

# creation_info is the only required property of the Document class (have a look there!), the rest are optional lists.
# So, we are set up to create a new document instance.
document = Document(creation_info)

# The document currently does not describe anything. Let's create a package that we can add to it.
# The Package class has quite a few properties (have a look there!),
# but only name, spdx_id and download_location are mandatory in SPDX v2.3.
package = Package(
    name="package name",
    spdx_id="SPDXRef-Package",
    download_location="https://download.com",
    version="2.2.1",
    file_name="./foo.bar",
    supplier=Actor(ActorType.PERSON, "Jane Doe", "jane.doe@example.com"),
    originator=Actor(ActorType.ORGANIZATION, "some organization", "contact@example.com"),
    files_analyzed=True,
    verification_code=PackageVerificationCode(
        value="d6a770ba38583ed4bb4525bd96e50461655d2758", excluded_files=["./some.file"]
    ),
    checksums=[
        Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2758"),
        Checksum(ChecksumAlgorithm.MD5, "624c1abb3664f4b35547e7c73864ad24"),
    ],
    license_concluded=spdx_licensing.parse("GPL-2.0-only OR MIT"),
    license_info_from_files=[spdx_licensing.parse("GPL-2.0-only"), spdx_licensing.parse("MIT")],
    license_declared=spdx_licensing.parse("GPL-2.0-only AND MIT"),
    license_comment="license comment",
    copyright_text="Copyright 2022 Jane Doe",
    description="package description",
    attribution_texts=["package attribution"],
    primary_package_purpose=PackagePurpose.LIBRARY,
    release_date=datetime(2015, 1, 1),
    external_references=[
        ExternalPackageRef(
            category=ExternalPackageRefCategory.OTHER,
            reference_type="http://reference.type",
            locator="reference/locator",
            comment="external reference comment",
        )
    ],
)

# Now that we have a package defined, we can add it to the document's package property.
document.packages = [package]

# A DESCRIBES relationship asserts that the document indeed describes the package.
describes_relationship = Relationship("SPDXRef-DOCUMENT", RelationshipType.DESCRIBES, "SPDXRef-Package")
document.relationships = [describes_relationship]

# Let's add two files. Have a look at the file class for all possible properties a file can have.
file1 = File(
    name="./package/file1.py",
    spdx_id="SPDXRef-File1",
    file_types=[FileType.SOURCE],
    checksums=[
        Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2758"),
        Checksum(ChecksumAlgorithm.MD5, "624c1abb3664f4b35547e7c73864ad24"),
    ],
    license_concluded=spdx_licensing.parse("MIT"),
    license_info_in_file=[spdx_licensing.parse("MIT")],
    copyright_text="Copyright 2022 Jane Doe",
)
file2 = File(
    name="./package/file2.py",
    spdx_id="SPDXRef-File2",
    checksums=[
        Checksum(ChecksumAlgorithm.SHA1, "d6a770ba38583ed4bb4525bd96e50461655d2759"),
    ],
    license_concluded=spdx_licensing.parse("GPL-2.0-only"),
)

# Assuming the package contains those two files, we create two CONTAINS relationships.
contains_relationship1 = Relationship("SPDXRef-Package", RelationshipType.CONTAINS, "SPDXRef-File1")
contains_relationship2 = Relationship("SPDXRef-Package", RelationshipType.CONTAINS, "SPDXRef-File2")

# This library uses run-time type checks when assigning properties.
# Because in-place alterations like .append() circumvent these checks, we don't use them here.
document.relationships += [contains_relationship1, contains_relationship2]
document.files += [file1, file2]

# We now have created a document with basic creation information, describing a package that contains two files.
# You can also add Annotations, Snippets and ExtractedLicensingInfo to the document in an analogous manner to the above.
# Have a look at their respective classes if you are unsure about their properties.


# This library provides comprehensive validation against the SPDX specification.
# Note that details of the validation depend on the SPDX version of the document.
validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)

# You can have a look at each entry's message and context (like spdx_id, parent_id, full_element)
# which will help you pinpoint the location of the invalidity.
for message in validation_messages:
    logging.warning(message.validation_message)
    logging.warning(message.context)

# If the document is valid, validation_messages will be empty.
assert validation_messages == []

# Finally, we can serialize the document to any of the five supported formats.
# Using the write_file() method from the write_anything module,
# the format will be determined by the file ending: .spdx (tag-value), .json, .xml, .yaml. or .rdf (or .rdf.xml)
write_file(document, "my_spdx_document.spdx.json")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os

import pytest

from spdx_tools.spdx3.validation.json_ld.shacl_validation import validate_against_shacl_from_file


@pytest.mark.skip("Currently the validation against SHACL fails, refer to process.md and the known limitations.")
def test_shacl_validation():
    # insert path to example json ld
    conforms, results_graph, results_text = validate_against_shacl_from_file(
        data_file=os.path.join(os.path.dirname(__file__), "../../../SPDX3_jsonld_test.jsonld"),
        shacl_file=os.path.join(
            os.path.dirname(__file__), "../../../../src/spdx_tools/spdx3/writer/json_ld/model.ttl"
        ),
    )
    # results_graph.serialize("validation_result.rdf.xml", format="pretty-xml")
    print(results_text)
    assert conforms



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

import pytest

from spdx_tools.spdx.validation.extracted_licensing_info_validator import validate_extracted_licensing_info
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import extracted_licensing_info_fixture


def test_valid_extracted_licensing_info():
    extracted_licensing_info = extracted_licensing_info_fixture()
    validation_messages: List[ValidationMessage] = validate_extracted_licensing_info(extracted_licensing_info)

    assert validation_messages == []


# TODO: tests for licenses not on the SPDX License list (i.e. they must provide id, name and cross-references)
@pytest.mark.parametrize(
    "extracted_licensing_info, expected_message",
    [
        (
            extracted_licensing_info_fixture(extracted_text=None),
            "extracted_text must be provided if there is a license_id assigned",
        ),
        (
            extracted_licensing_info_fixture(cross_references=["invalid_url"]),
            "cross_reference must be a valid URL, but is: invalid_url",
        ),
    ],
)
def test_invalid_extracted_licensing_info(extracted_licensing_info, expected_message):
    validation_messages: List[ValidationMessage] = validate_extracted_licensing_info(extracted_licensing_info)

    expected = ValidationMessage(
        expected_message,
        ValidationContext(
            element_type=SpdxElementType.EXTRACTED_LICENSING_INFO, full_element=extracted_licensing_info
        ),
    )

    assert validation_messages == [expected]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx3.bump_from_spdx2.package import bump_package
from spdx_tools.spdx3.model import ExternalIdentifier, ExternalIdentifierType, ExternalReference, ExternalReferenceType
from spdx_tools.spdx3.model.software import Package
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import SpdxNoAssertion
from spdx_tools.spdx.model.package import ExternalPackageRef, ExternalPackageRefCategory
from spdx_tools.spdx.model.package import Package as Spdx2_Package
from tests.spdx.fixtures import actor_fixture, package_fixture


@pytest.mark.parametrize(
    "originator, expected_originator, supplier, expected_supplier",
    [
        (
            actor_fixture(name="originatorName"),
            ["https://doc.namespace#SPDXRef-Actor-originatorName-some@mail.com"],
            actor_fixture(name="supplierName"),
            ["https://doc.namespace#SPDXRef-Actor-supplierName-some@mail.com"],
        ),
        (None, [], None, []),
        (SpdxNoAssertion(), [], SpdxNoAssertion(), []),
    ],
)
def test_bump_package(originator, expected_originator, supplier, expected_supplier):
    payload = Payload()
    document_namespace = "https://doc.namespace"
    spdx2_package: Spdx2_Package = package_fixture(
        originator=originator,
        supplier=supplier,
        external_references=[
            ExternalPackageRef(
                ExternalPackageRefCategory.SECURITY, "advisory", "advisory_locator", "advisory_comment"
            ),
            ExternalPackageRef(ExternalPackageRefCategory.PERSISTENT_ID, "swh", "swh_locator", "swh_comment"),
        ],
    )
    expected_new_package_id = f"{document_namespace}#{spdx2_package.spdx_id}"

    bump_package(spdx2_package, payload, document_namespace, [], [])
    package = payload.get_element(expected_new_package_id)

    assert isinstance(package, Package)
    assert package.spdx_id == expected_new_package_id
    assert package.package_version == spdx2_package.version
    assert package.external_reference == [
        ExternalReference(ExternalReferenceType.SECURITY_ADVISORY, ["advisory_locator"], None, "advisory_comment")
    ]
    assert package.external_identifier == [
        ExternalIdentifier(ExternalIdentifierType.SWHID, "swh_locator", "swh_comment")
    ]
    assert package.download_location == spdx2_package.download_location
    assert package.package_version == spdx2_package.version
    assert package.originated_by == expected_originator
    assert package.supplied_by == expected_supplier
    assert package.homepage == spdx2_package.homepage
    assert package.source_info == spdx2_package.source_info
    assert package.built_time == spdx2_package.built_date
    assert package.release_time == spdx2_package.release_date
    assert package.valid_until_time == spdx2_package.valid_until_date
    assert package.copyright_text == spdx2_package.copyright_text
    assert package.attribution_text == spdx2_package.attribution_texts[0]


def test_bump_of_single_purl_without_comment():
    payload = Payload()
    document_namespace = "https://doc.namespace"
    spdx2_package: Spdx2_Package = package_fixture(
        external_references=[
            ExternalPackageRef(ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "purl_locator", None),
        ]
    )
    expected_new_package_id = f"{document_namespace}#{spdx2_package.spdx_id}"

    bump_package(spdx2_package, payload, document_namespace, [], [])
    package = payload.get_element(expected_new_package_id)

    assert package.package_url == "purl_locator"
    assert package.external_reference == []
    assert package.external_identifier == []


def test_bump_of_single_purl_with_comment():
    payload = Payload()
    document_namespace = "https://doc.namespace"
    spdx2_package: Spdx2_Package = package_fixture(
        external_references=[
            ExternalPackageRef(ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "purl_locator", "purl_comment"),
        ]
    )
    expected_new_package_id = f"{document_namespace}#{spdx2_package.spdx_id}"

    bump_package(spdx2_package, payload, document_namespace, [], [])
    package = payload.get_element(expected_new_package_id)

    assert package.package_url is None
    assert package.external_reference == []
    assert package.external_identifier == [
        ExternalIdentifier(ExternalIdentifierType.PURL, "purl_locator", "purl_comment")
    ]


def test_bump_of_multiple_purls():
    payload = Payload()
    document_namespace = "https://doc.namespace"
    spdx2_package: Spdx2_Package = package_fixture(
        external_references=[
            ExternalPackageRef(ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "purl_locator", "comment"),
            ExternalPackageRef(ExternalPackageRefCategory.PACKAGE_MANAGER, "purl", "purl_locator2", None),
        ]
    )
    expected_new_package_id = f"{document_namespace}#{spdx2_package.spdx_id}"

    bump_package(spdx2_package, payload, document_namespace, [], [])
    package = payload.get_element(expected_new_package_id)

    assert package.package_url is None
    assert package.external_reference == []
    TestCase().assertCountEqual(
        package.external_identifier,
        [
            ExternalIdentifier(ExternalIdentifierType.PURL, "purl_locator", "comment"),
            ExternalIdentifier(ExternalIdentifierType.PURL, "purl_locator2", None),
        ],
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import logging
from datetime import datetime

from beartype.typing import Any, Dict, Optional
from rdflib import Graph, Literal
from rdflib.term import Node

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.validation.spdx_id_validators import is_valid_internal_spdx_id


def add_optional_literal(value: Any, graph: Graph, parent: Node, predicate: Node):
    if value is None:
        return
    if isinstance(value, list):
        for element in value:
            graph.add((parent, predicate, Literal(str(element))))
        return
    graph.add((parent, predicate, Literal(str(value))))


def add_literal_or_no_assertion_or_none(value: Any, graph: Graph, parent: Node, predicate: Node):
    if value is None:
        return
    if isinstance(value, SpdxNone):
        graph.add((parent, predicate, SPDX_NAMESPACE.none))
        return
    add_literal_or_no_assertion(value, graph, parent, predicate)


def add_literal_or_no_assertion(value: Any, graph: Graph, parent: Node, predicate: Node):
    if value is None:
        return
    if isinstance(value, SpdxNoAssertion):
        graph.add((parent, predicate, SPDX_NAMESPACE.noassertion))
        return
    add_optional_literal(value, graph, parent, predicate)


def add_datetime_to_graph(value: Optional[datetime], graph: Graph, parent: Node, predicate: Node):
    if value:
        graph.add((parent, predicate, Literal(datetime_to_iso_string(value))))


def add_namespace_to_spdx_id(spdx_id: str, doc_namespace: str, external_doc_namespaces: Dict[str, str]) -> str:
    if ":" in spdx_id:
        external_doc_ref_id = spdx_id.split(":")[0]
        if external_doc_ref_id not in external_doc_namespaces.keys():
            logging.warning(f"No namespace for external document reference with id {external_doc_ref_id} provided.")
            return spdx_id
        return f"{external_doc_namespaces[external_doc_ref_id]}#{spdx_id.split(':')[1]}"

    if is_valid_internal_spdx_id(spdx_id):
        return f"{doc_namespace}#{spdx_id}"

    return spdx_id



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Union
from rdflib import RDFS, BNode, Graph, URIRef

from spdx_tools.spdx.model import File, FileType
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.checksum_parser import parse_checksum
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import (
    apply_parsing_method_or_log_error,
    get_correctly_typed_triples,
    get_correctly_typed_value,
    parse_enum_value,
    parse_literal,
    parse_literal_or_no_assertion_or_none,
    parse_spdx_id,
)
from spdx_tools.spdx.parser.rdf.license_expression_parser import parse_license_expression
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def parse_file(file_node: Union[URIRef, BNode], graph: Graph, doc_namespace: str) -> File:
    logger = Logger()
    spdx_id = parse_spdx_id(file_node, doc_namespace, graph)
    name = parse_literal(logger, graph, file_node, SPDX_NAMESPACE.fileName)
    checksums = []
    for _, _, checksum_node in get_correctly_typed_triples(logger, graph, file_node, SPDX_NAMESPACE.checksum):
        checksums.append(parse_checksum(checksum_node, graph))

    file_types = []
    for _, _, file_type_ref in graph.triples((file_node, SPDX_NAMESPACE.fileType, None)):
        file_types.append(
            apply_parsing_method_or_log_error(
                logger, file_type_ref, parsing_method=lambda x: parse_enum_value(x, FileType, SPDX_NAMESPACE.fileType_)
            )
        )
    license_concluded = parse_literal_or_no_assertion_or_none(
        logger,
        graph,
        file_node,
        SPDX_NAMESPACE.licenseConcluded,
        parsing_method=lambda x: parse_license_expression(x, graph, doc_namespace, logger),
    )
    license_info_in_file = []
    for _, _, license_info_from_files_node in graph.triples((file_node, SPDX_NAMESPACE.licenseInfoInFile, None)):
        license_info_in_file.append(
            get_correctly_typed_value(
                logger,
                license_info_from_files_node,
                lambda x: parse_license_expression(x, graph, doc_namespace, logger),
            )
        )
    license_comment = parse_literal(logger, graph, file_node, SPDX_NAMESPACE.licenseComments)
    copyright_text = parse_literal_or_no_assertion_or_none(logger, graph, file_node, SPDX_NAMESPACE.copyrightText)
    file_contributors = []
    for _, _, file_contributor in get_correctly_typed_triples(
        logger, graph, file_node, SPDX_NAMESPACE.fileContributor, None
    ):
        file_contributors.append(file_contributor.toPython())

    notice_text = parse_literal(logger, graph, file_node, SPDX_NAMESPACE.noticeText)
    comment = parse_literal(logger, graph, file_node, RDFS.comment)
    attribution_texts = []
    for _, _, attribution_text_literal in get_correctly_typed_triples(
        logger, graph, file_node, SPDX_NAMESPACE.attributionText, None
    ):
        attribution_texts.append(attribution_text_literal.toPython())
    raise_parsing_error_if_logger_has_messages(logger, "File")
    file = construct_or_raise_parsing_error(
        File,
        dict(
            name=name,
            spdx_id=spdx_id,
            checksums=checksums,
            attribution_texts=attribution_texts,
            comment=comment,
            copyright_text=copyright_text,
            file_types=file_types,
            contributors=file_contributors,
            license_comment=license_comment,
            license_concluded=license_concluded,
            license_info_in_file=license_info_in_file,
            notice=notice_text,
        ),
    )
    return file



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List


class ConstructorTypeErrors(TypeError):
    """
    Helper class that holds a list of error messages. Intended to capture all TypeErrors encountered during a
    constructor call, instead of raising only the first one.
    """

    messages: List[str]

    def __init__(self, messages: List[str]):
        self.messages = messages

    def get_messages(self):
        return list(self.messages)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import ActorType
from spdx_tools.spdx.validation.actor_validator import validate_actor
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import actor_fixture


def test_valid_actor_person():
    actor = actor_fixture()
    validation_messages: List[ValidationMessage] = validate_actor(actor, DOCUMENT_SPDX_ID)

    assert validation_messages == []


@pytest.mark.parametrize(
    "actor, expected_message",
    [
        (
            actor_fixture(actor_type=ActorType.TOOL, email="mail@mail.com"),
            "email must be None if actor_type is TOOL, but is: mail@mail.com",
        ),
    ],
)
def test_invalid_actor(actor, expected_message):
    parent_id = DOCUMENT_SPDX_ID
    validation_messages: List[ValidationMessage] = validate_actor(actor, parent_id)

    expected = ValidationMessage(
        expected_message,
        ValidationContext(parent_id=parent_id, element_type=SpdxElementType.ACTOR, full_element=actor),
    )

    assert validation_messages == [expected]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from datetime import datetime

from beartype.typing import Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.spdx3.model import Relationship


@dataclass_with_properties
class VulnAssessmentRelationship(Relationship):
    assessed_element: Optional[str] = None  # id of the element
    published_time: Optional[datetime] = None
    supplied_by: Optional[str] = None
    modified_time: Optional[datetime] = None
    withdrawn_time: Optional[datetime] = None

    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Union
from unittest import mock
from unittest.mock import MagicMock, NonCallableMagicMock

import pytest

from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.document_converter import DocumentConverter
from spdx_tools.spdx.jsonschema.document_properties import DocumentProperty
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Annotation,
    AnnotationType,
    Document,
    ExtractedLicensingInfo,
    Relationship,
    RelationshipType,
)
from tests.spdx.fixtures import (
    annotation_fixture,
    creation_info_fixture,
    external_document_ref_fixture,
    file_fixture,
    package_fixture,
    snippet_fixture,
)
from tests.spdx.mock_utils import assert_mock_method_called_with_arguments


@pytest.fixture
@mock.patch("spdx_tools.spdx.jsonschema.creation_info_converter.CreationInfoConverter", autospec=True)
@mock.patch("spdx_tools.spdx.jsonschema.external_document_ref_converter.ExternalDocumentRefConverter", autospec=True)
@mock.patch("spdx_tools.spdx.jsonschema.package_converter.PackageConverter", autospec=True)
@mock.patch("spdx_tools.spdx.jsonschema.annotation_converter.AnnotationConverter", autospec=True)
@mock.patch(
    "spdx_tools.spdx.jsonschema.extracted_licensing_info_converter.ExtractedLicensingInfoConverter", autospec=True
)
@mock.patch("spdx_tools.spdx.jsonschema.file_converter.FileConverter", autospec=True)
@mock.patch("spdx_tools.spdx.jsonschema.snippet_converter.SnippetConverter", autospec=True)
@mock.patch("spdx_tools.spdx.jsonschema.relationship_converter.RelationshipConverter", autospec=True)
def converter(
    relationship_converter_mock: MagicMock,
    snippet_converter_mock: MagicMock,
    file_converter_mock: MagicMock,
    extracted_licensing_info_converter_mock: MagicMock,
    annotation_converter_mock: MagicMock,
    package_converter_mock: MagicMock,
    external_ref_converter_mock: MagicMock,
    creation_info_converter_mock: MagicMock,
) -> DocumentConverter:
    converter = DocumentConverter()
    converter.creation_info_converter = creation_info_converter_mock()
    converter.external_document_ref_converter = external_ref_converter_mock()
    converter.package_converter = package_converter_mock()
    converter.annotation_converter = annotation_converter_mock()
    converter.extracted_licensing_info_converter = extracted_licensing_info_converter_mock()
    converter.file_converter = file_converter_mock()
    converter.snippet_converter = snippet_converter_mock()
    converter.relationship_converter = relationship_converter_mock()
    return converter


@pytest.mark.parametrize(
    "document_property,expected",
    [
        (DocumentProperty.SPDX_VERSION, "spdxVersion"),
        (DocumentProperty.SPDX_ID, "SPDXID"),
        (DocumentProperty.NAME, "name"),
        (DocumentProperty.DOCUMENT_NAMESPACE, "documentNamespace"),
        (DocumentProperty.DATA_LICENSE, "dataLicense"),
        (DocumentProperty.EXTERNAL_DOCUMENT_REFS, "externalDocumentRefs"),
        (DocumentProperty.COMMENT, "comment"),
        (DocumentProperty.CREATION_INFO, "creationInfo"),
        (DocumentProperty.PACKAGES, "packages"),
        (DocumentProperty.FILES, "files"),
        (DocumentProperty.SNIPPETS, "snippets"),
        (DocumentProperty.ANNOTATIONS, "annotations"),
        (DocumentProperty.RELATIONSHIPS, "relationships"),
        (DocumentProperty.HAS_EXTRACTED_LICENSING_INFOS, "hasExtractedLicensingInfos"),
    ],
)
def test_json_property_names(converter: DocumentConverter, document_property: DocumentProperty, expected: str):
    assert converter.json_property_name(document_property) == expected


def test_successful_conversion(converter: DocumentConverter):
    creation_info = creation_info_fixture(
        spdx_version="spdxVersion",
        spdx_id="spdxId",
        name="name",
        document_namespace="namespace",
        document_comment="comment",
        data_license="dataLicense",
        external_document_refs=[external_document_ref_fixture()],
    )
    document = Document(
        creation_info,
        annotations=[
            Annotation(
                "annotationId",
                AnnotationType.REVIEW,
                Actor(ActorType.PERSON, "reviewerName"),
                datetime(2022, 12, 1),
                "reviewComment",
            )
        ],
        extracted_licensing_info=[ExtractedLicensingInfo("licenseId", "licenseText")],
        relationships=[
            Relationship(creation_info.spdx_id, RelationshipType.DESCRIBES, "describedElementId"),
            Relationship("relationshipOriginId", RelationshipType.AMENDS, "relationShipTargetId"),
        ],
        packages=[package_fixture()],
        files=[file_fixture()],
        snippets=[snippet_fixture()],
    )
    converter.external_document_ref_converter.convert.return_value = "mock_converted_external_ref"
    converter.creation_info_converter.convert.return_value = "mock_converted_creation_info"
    converter.package_converter.convert.return_value = "mock_converted_package"
    converter.annotation_converter.convert.return_value = "mock_converted_annotation"
    converter.extracted_licensing_info_converter.convert.return_value = "mock_converted_extracted_licensing_info"
    converter.package_converter.convert.return_value = "mock_converted_package"
    converter.file_converter.convert.return_value = "mock_converted_file"
    converter.snippet_converter.convert.return_value = "mock_converted_snippet"
    converter.relationship_converter.convert.return_value = "mock_converted_relationship"

    converted_dict = converter.convert(document)

    assert converted_dict == {
        converter.json_property_name(DocumentProperty.SPDX_ID): "spdxId",
        converter.json_property_name(DocumentProperty.ANNOTATIONS): ["mock_converted_annotation"],
        converter.json_property_name(DocumentProperty.COMMENT): "comment",
        converter.json_property_name(DocumentProperty.CREATION_INFO): "mock_converted_creation_info",
        converter.json_property_name(DocumentProperty.DATA_LICENSE): "dataLicense",
        converter.json_property_name(DocumentProperty.EXTERNAL_DOCUMENT_REFS): ["mock_converted_external_ref"],
        converter.json_property_name(DocumentProperty.HAS_EXTRACTED_LICENSING_INFOS): [
            "mock_converted_extracted_licensing_info"
        ],
        converter.json_property_name(DocumentProperty.NAME): "name",
        converter.json_property_name(DocumentProperty.SPDX_VERSION): "spdxVersion",
        converter.json_property_name(DocumentProperty.DOCUMENT_NAMESPACE): "namespace",
        converter.json_property_name(DocumentProperty.PACKAGES): ["mock_converted_package"],
        converter.json_property_name(DocumentProperty.FILES): ["mock_converted_file"],
        converter.json_property_name(DocumentProperty.SNIPPETS): ["mock_converted_snippet"],
        converter.json_property_name(DocumentProperty.RELATIONSHIPS): [
            "mock_converted_relationship",
            "mock_converted_relationship",
        ],
    }


def test_json_type(converter: DocumentConverter):
    assert converter.get_json_type() == DocumentProperty


def test_data_model_type(converter: DocumentConverter):
    assert converter.get_data_model_type() == Document


def test_null_values(converter: DocumentConverter):
    document = Document(creation_info_fixture(external_document_refs=[]))

    converted_dict = converter.convert(document)

    assert converter.json_property_name(DocumentProperty.ANNOTATIONS) not in converted_dict
    assert converter.json_property_name(DocumentProperty.EXTERNAL_DOCUMENT_REFS) not in converted_dict
    assert converter.json_property_name(DocumentProperty.HAS_EXTRACTED_LICENSING_INFOS) not in converted_dict
    assert converter.json_property_name(DocumentProperty.PACKAGES) not in converted_dict
    assert converter.json_property_name(DocumentProperty.FILES) not in converted_dict
    assert converter.json_property_name(DocumentProperty.SNIPPETS) not in converted_dict
    assert converter.json_property_name(DocumentProperty.RELATIONSHIPS) not in converted_dict


def test_document_annotations(converter: DocumentConverter):
    file = file_fixture(spdx_id="fileId")
    package = package_fixture(spdx_id="packageId")
    snippet = snippet_fixture(spdx_id="snippetId")
    document_id = "documentId"

    # There are 5 annotations: one each referencing the document, package, file and snippet, and one with an id
    # matching none of the Spdx elements. The writer is expected to add the package, file and snippet annotations to
    # those elements, so the document should receive the other two.
    document_annotation = annotation_fixture(spdx_id=document_id)
    other_annotation = annotation_fixture(spdx_id="otherId")
    annotations = [
        annotation_fixture(spdx_id=file.spdx_id),
        annotation_fixture(spdx_id=package.spdx_id),
        annotation_fixture(spdx_id=snippet.spdx_id),
        document_annotation,
        other_annotation,
    ]
    document = Document(
        creation_info_fixture(spdx_id=document_id),
        files=[file],
        packages=[package],
        snippets=[snippet],
        annotations=annotations,
    )

    # Weird type hint to make warnings about unresolved references from the mock class disappear
    annotation_converter: Union[AnnotationConverter, NonCallableMagicMock] = converter.annotation_converter
    annotation_converter.convert.return_value = "mock_converted_annotation"

    converted_dict = converter.convert(document)

    assert_mock_method_called_with_arguments(annotation_converter, "convert", document_annotation, other_annotation)
    converted_document_annotations = converted_dict.get(converter.json_property_name(DocumentProperty.ANNOTATIONS))
    assert converted_document_annotations == ["mock_converted_annotation", "mock_converted_annotation"]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List

from spdx_tools.spdx3.bump_from_spdx2.checksum import bump_checksum
from spdx_tools.spdx3.bump_from_spdx2.message import print_missing_conversion
from spdx_tools.spdx3.model import ExternalMap
from spdx_tools.spdx3.model.software import File
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import ExternalDocumentRef, SpdxNoAssertion
from spdx_tools.spdx.model.file import File as Spdx2_File
from spdx_tools.spdx.spdx_element_utils import get_full_element_spdx_id


def bump_file(
    spdx2_file: Spdx2_File,
    payload: Payload,
    document_namespace: str,
    external_document_refs: List[ExternalDocumentRef],
    imports: List[ExternalMap],
):
    spdx_id = get_full_element_spdx_id(spdx2_file, document_namespace, external_document_refs)
    if ":" in spdx2_file.spdx_id:
        imports.append(
            ExternalMap(
                external_id=spdx2_file.spdx_id,
                defining_document=f"{spdx2_file.spdx_id.split(':')[0]}:SPDXRef-DOCUMENT",
            )
        )

    integrity_methods = [bump_checksum(checksum) for checksum in spdx2_file.checksums]
    print_missing_conversion(
        "file.file_type", 0, "different cardinalities, " "https://github.com/spdx/spdx-3-model/issues/82"
    )
    copyright_text = None
    if isinstance(spdx2_file.copyright_text, str):
        copyright_text = spdx2_file.copyright_text
    elif isinstance(spdx2_file.copyright_text, SpdxNoAssertion):
        print_missing_conversion("package2.copyright_text", 0)
    print_missing_conversion(
        "file.notice, file.contributors, file.license_info_in_file, file.license_comment",
        0,
        "missing definition for license profile",
    )

    payload.add_element(
        File(
            spdx_id,
            name=spdx2_file.name,
            comment=spdx2_file.comment,
            verified_using=integrity_methods,
            copyright_text=copyright_text,
            attribution_text=", ".join(spdx2_file.attribution_texts),
        )
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from datetime import datetime

from beartype.typing import Dict, List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import CreationInfo, Element, ExternalIdentifier, ExternalReference, Hash, IntegrityMethod


@dataclass_with_properties
class Build(Element):
    build_type: str = None
    build_id: Optional[str] = None
    config_source_entrypoint: List[str] = field(default_factory=list)
    config_source_uri: List[str] = field(default_factory=list)
    config_source_digest: List[Hash] = field(default_factory=list)
    parameters: Dict[str, str] = field(default_factory=dict)
    build_start_time: Optional[datetime] = None
    build_end_time: Optional[datetime] = None
    environment: Dict[str, str] = field(default_factory=dict)

    def __init__(
        self,
        spdx_id: str,
        build_type: str,
        creation_info: Optional[CreationInfo] = None,
        name: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        verified_using: List[IntegrityMethod] = None,
        external_reference: List[ExternalReference] = None,
        external_identifier: List[ExternalIdentifier] = None,
        extension: Optional[str] = None,
        build_id: Optional[str] = None,
        config_source_entrypoint: List[str] = None,
        config_source_uri: List[str] = None,
        config_source_digest: List[Hash] = None,
        parameters: Dict[str, str] = None,
        build_start_time: Optional[datetime] = None,
        build_end_time: Optional[datetime] = None,
        environment: Dict[str, str] = None,
    ):
        verified_using = [] if verified_using is None else verified_using
        external_reference = [] if external_reference is None else external_reference
        external_identifier = [] if external_identifier is None else external_identifier
        config_source_entrypoint = [] if config_source_entrypoint is None else config_source_entrypoint
        config_source_uri = [] if config_source_uri is None else config_source_uri
        config_source_digest = [] if config_source_digest is None else config_source_digest
        parameters = {} if parameters is None else parameters
        environment = {} if environment is None else environment

        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Tuple

from spdx_tools.spdx3.bump_from_spdx2.checksum import bump_checksum
from spdx_tools.spdx3.model import ExternalMap, Hash, NamespaceMap
from spdx_tools.spdx.model.external_document_ref import ExternalDocumentRef


def bump_external_document_ref(external_document_ref: ExternalDocumentRef) -> Tuple[NamespaceMap, ExternalMap]:
    verified_using: List[Hash] = [bump_checksum(external_document_ref.checksum)]

    return NamespaceMap(external_document_ref.document_ref_id, external_document_ref.document_uri + "#"), ExternalMap(
        external_id=f"{external_document_ref.document_ref_id}:SPDXRef-DOCUMENT",
        verified_using=verified_using,
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest
from rdflib import DOAP, RDF, RDFS, XSD, Graph, Literal, URIRef

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.model import ExternalPackageRefCategory
from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.package_writer import (
    add_external_package_ref_to_graph,
    add_package_to_graph,
    add_package_verification_code_to_graph,
)
from tests.spdx.fixtures import external_package_ref_fixture, package_fixture, package_verification_code_fixture


def test_add_package_to_graph():
    graph = Graph()
    package = package_fixture()

    add_package_to_graph(package, graph, "docNamespace", {})

    assert (URIRef("docNamespace#SPDXRef-Package"), RDF.type, SPDX_NAMESPACE.Package) in graph
    assert (None, SPDX_NAMESPACE.name, Literal(package.name)) in graph
    assert (None, SPDX_NAMESPACE.versionInfo, Literal(package.version)) in graph
    assert (None, SPDX_NAMESPACE.packageFileName, Literal(package.file_name)) in graph
    assert (None, SPDX_NAMESPACE.supplier, Literal(package.supplier.to_serialized_string())) in graph
    assert (None, SPDX_NAMESPACE.originator, Literal(package.originator.to_serialized_string())) in graph
    assert (None, SPDX_NAMESPACE.downloadLocation, Literal(package.download_location)) in graph
    assert (None, SPDX_NAMESPACE.filesAnalyzed, Literal(package.files_analyzed, datatype=XSD.boolean)) in graph
    assert (URIRef("docNamespace#SPDXRef-Package"), SPDX_NAMESPACE.packageVerificationCode, None) in graph
    assert (URIRef("docNamespace#SPDXRef-Package"), SPDX_NAMESPACE.checksum, None) in graph
    assert (None, DOAP.homepage, Literal(package.homepage)) in graph
    assert (None, SPDX_NAMESPACE.sourceInfo, Literal(package.source_info)) in graph
    assert (None, SPDX_NAMESPACE.licenseConcluded, None) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoFromFiles, LICENSE_NAMESPACE.MIT) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoFromFiles, LICENSE_NAMESPACE["GPL-2.0-only"]) in graph
    assert (None, SPDX_NAMESPACE.licenseInfoFromFiles, SPDX_NAMESPACE.noassertion) in graph
    assert (None, SPDX_NAMESPACE.licenseDeclared, None) in graph
    assert (None, SPDX_NAMESPACE.licenseComments, Literal(package.license_comment)) in graph
    assert (None, SPDX_NAMESPACE.copyrightText, Literal(package.copyright_text)) in graph
    assert (None, SPDX_NAMESPACE.summary, Literal(package.summary)) in graph
    assert (None, SPDX_NAMESPACE.description, Literal(package.description)) in graph
    assert (None, RDFS.comment, Literal(package.comment)) in graph
    assert (URIRef("docNamespace#SPDXRef-Package"), SPDX_NAMESPACE.externalRef, None) in graph
    assert (None, SPDX_NAMESPACE.attributionText, Literal(package.attribution_texts[0])) in graph
    assert (None, SPDX_NAMESPACE.primaryPackagePurpose, SPDX_NAMESPACE.purpose_source) in graph
    assert (None, SPDX_NAMESPACE.releaseDate, Literal(datetime_to_iso_string(package.release_date))) in graph
    assert (None, SPDX_NAMESPACE.builtDate, Literal(datetime_to_iso_string(package.built_date))) in graph
    assert (None, SPDX_NAMESPACE.validUntilDate, Literal(datetime_to_iso_string(package.valid_until_date))) in graph


def test_add_package_verification_code_to_graph():
    graph = Graph()
    verification_code = package_verification_code_fixture()

    add_package_verification_code_to_graph(verification_code, graph, URIRef("docNamespace"))

    assert (None, RDF.type, SPDX_NAMESPACE.PackageVerificationCode) in graph
    assert (
        None,
        SPDX_NAMESPACE.packageVerificationCodeValue,
        Literal("85ed0817af83a24ad8da68c2b5094de69833983c"),
    ) in graph
    assert (None, SPDX_NAMESPACE.packageVerificationCodeExcludedFile, Literal("./exclude.py")) in graph


@pytest.mark.parametrize(
    "external_reference,ref_type,category",
    [
        (
            external_package_ref_fixture(),
            URIRef("http://spdx.org/rdf/references/maven-central"),
            SPDX_NAMESPACE.referenceCategory_packageManager,
        ),
        (
            external_package_ref_fixture(
                locator="acmecorp/acmenator/4.1.3-alpha",
                category=ExternalPackageRefCategory.OTHER,
                reference_type="LocationRef-acmeforge",
                comment="This is the external ref for Acme",
            ),
            URIRef("https://some.namespace#LocationRef-acmeforge"),
            SPDX_NAMESPACE.referenceCategory_other,
        ),
    ],
)
def test_external_package_ref_to_graph(external_reference, ref_type, category):
    graph = Graph()
    doc_namespace = "https://some.namespace"
    add_external_package_ref_to_graph(external_reference, graph, URIRef("docNamespace"), doc_namespace)

    assert (None, RDF.type, SPDX_NAMESPACE.ExternalRef) in graph
    assert (None, SPDX_NAMESPACE.referenceCategory, category) in graph
    assert (None, SPDX_NAMESPACE.referenceType, ref_type) in graph
    assert (None, SPDX_NAMESPACE.referenceLocator, Literal(external_reference.locator)) in graph
    assert (None, RDFS.comment, Literal(external_reference.comment)) in graph



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List

from spdx_tools.spdx.model import Actor, ActorType
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_actors(actors: List[Actor], parent_id: str) -> List[ValidationMessage]:
    validation_messages = []
    for actor in actors:
        validation_messages.extend(validate_actor(actor, parent_id))

    return validation_messages


def validate_actor(actor: Actor, parent_id: str) -> List[ValidationMessage]:
    validation_messages = []

    if actor.actor_type == ActorType.TOOL and actor.email is not None:
        validation_messages.append(
            ValidationMessage(
                f"email must be None if actor_type is TOOL, but is: {actor.email}",
                ValidationContext(parent_id=parent_id, element_type=SpdxElementType.ACTOR, full_element=actor),
            )
        )

    return validation_messages



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import os

import pytest

from spdx_tools.spdx.model import Document
from spdx_tools.spdx.parser.json import json_parser
from spdx_tools.spdx.parser.rdf import rdf_parser
from spdx_tools.spdx.parser.tagvalue import tagvalue_parser
from spdx_tools.spdx.parser.xml import xml_parser
from spdx_tools.spdx.parser.yaml import yaml_parser


@pytest.mark.parametrize(
    "parser, format_name, extension",
    [
        (json_parser, "JSON", ".json"),
        (xml_parser, "XML", ".xml"),
        (yaml_parser, "YAML", ".yaml"),
        (rdf_parser, "Rdf", ".rdf.xml"),
        (tagvalue_parser, "Tag", ""),
    ],
)
class TestParseFromFile:
    def test_parse_from_file_not_found(self, parser, format_name, extension):
        with pytest.raises(FileNotFoundError) as err:
            wrong_file_path = os.path.join(os.path.dirname(__file__), f"hnjfkjsedhnflsiafg.spdx{extension}")
            parser.parse_from_file(wrong_file_path)

        assert err.value.args[1] == "No such file or directory"

    def test_parse_from_file_with_2_3_example(self, parser, format_name, extension):
        doc = parser.parse_from_file(
            os.path.join(os.path.dirname(__file__), f"../../data/SPDX{format_name}Example-v2.3.spdx{extension}")
        )
        assert isinstance(doc, Document)
        assert len(doc.annotations) == 5
        assert len(doc.files) == 5
        assert len(doc.packages) == 4
        assert len(doc.snippets) == 1
        assert len(doc.relationships) == 13
        assert len(doc.extracted_licensing_info) == 5

    def test_parse_from_file_with_2_2_example(self, parser, format_name, extension):
        doc = parser.parse_from_file(
            os.path.join(os.path.dirname(__file__), f"../../data/SPDX{format_name}Example-v2.2.spdx{extension}")
        )
        assert isinstance(doc, Document)
        assert len(doc.annotations) == 5
        assert len(doc.files) == 4
        assert len(doc.packages) == 4
        assert len(doc.snippets) == 1
        assert len(doc.relationships) == 11
        assert len(doc.extracted_licensing_info) == 5

    def test_parse_from_file_with_encoding_example(self, parser, format_name, extension):
        doc = parser.parse_from_file(
            os.path.join(os.path.dirname(__file__), f"../../data/SPDX{format_name}Example-UTF-16.spdx{extension}"),
            "utf-16",
        )
        assert isinstance(doc, Document)
        assert len(doc.annotations) == 5
        assert len(doc.files) == 4
        assert len(doc.packages) == 4
        assert len(doc.snippets) == 1
        assert len(doc.relationships) == 11
        assert len(doc.extracted_licensing_info) == 5



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm


def test_correct_initialization():
    checksum = Checksum(ChecksumAlgorithm.BLAKE2B_256, "value")
    assert checksum.algorithm == ChecksumAlgorithm.BLAKE2B_256
    assert checksum.value == "value"


def test_wrong_type_in_algorithm():
    with pytest.raises(TypeError):
        Checksum(42, "value")


def test_wrong_type_in_value():
    with pytest.raises(TypeError):
        Checksum(ChecksumAlgorithm.BLAKE2B_256, 42)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest import TestCase

import pytest

from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser
from tests.spdx.parser.tagvalue.test_creation_info_parser import DOCUMENT_STR


def test_parse_extracted_licensing_info():
    parser = Parser()
    extracted_licensing_info_str = "\n".join(
        [
            "LicenseID: LicenseRef-Beerware-4.2",
            'ExtractedText: <text>"THE BEER-WARE LICENSE" (Revision 42): phk@FreeBSD.ORG wrote this file. As long as'
            " you retain this notice you can do whatever you want with this stuff. If we meet some day, and you think "
            "this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp</text>",
            "LicenseName: Beer-Ware License (Version 42)",
            "LicenseCrossReference:  http://people.freebsd.org/~phk/",
            "LicenseCrossReference:  http://another.cross.reference/",
            "LicenseComment: The beerware license has a couple of other standard variants.",
        ]
    )
    document = parser.parse("\n".join([DOCUMENT_STR, extracted_licensing_info_str]))
    assert document is not None
    assert len(document.extracted_licensing_info) == 1
    extracted_licensing_info = document.extracted_licensing_info[0]
    assert extracted_licensing_info.license_id == "LicenseRef-Beerware-4.2"
    assert (
        extracted_licensing_info.extracted_text
        == '"THE BEER-WARE LICENSE" (Revision 42): phk@FreeBSD.ORG wrote this file. '
        "As long as you retain this notice you can do whatever you want with this stuff. "
        "If we meet some day, and you think this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp"
    )
    assert extracted_licensing_info.license_name == "Beer-Ware License (Version 42)"
    TestCase().assertCountEqual(
        extracted_licensing_info.cross_references,
        ["http://people.freebsd.org/~phk/", "http://another.cross.reference/"],
    )
    assert extracted_licensing_info.comment == "The beerware license has a couple of other standard variants."


def test_parse_invalid_extracted_licensing_info():
    parser = Parser()
    extracted_licensing_info_str = "\n".join(
        [
            'ExtractedText: <text>"THE BEER-WARE LICENSE" (Revision 42): phk@FreeBSD.ORG wrote this file. '
            "As long as you retain this notice you can do whatever you want with this stuff. If we meet some day, "
            "and you think this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp</text>",
            "LicenseName: Beer-Ware License (Version 42)",
            "LicenseCrossReference:  http://people.freebsd.org/~phk/",
            "LicenseComment: The beerware license has a couple of other standard variants.",
        ]
    )

    with pytest.raises(SPDXParsingError) as err:
        parser.parse(extracted_licensing_info_str)

    assert err.value.get_messages() == [
        "Element ExtractedLicensingInfo is not the current element in scope, probably "
        "the expected tag to start the element (LicenseID) is missing. Line: 1",
        "Element ExtractedLicensingInfo is not the current element in scope, probably "
        "the expected tag to start the element (LicenseID) is missing. Line: 2",
        "Element ExtractedLicensingInfo is not the current element in scope, probably "
        "the expected tag to start the element (LicenseID) is missing. Line: 3",
        "Element ExtractedLicensingInfo is not the current element in scope, probably "
        "the expected tag to start the element (LicenseID) is missing. Line: 4",
    ]



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto
from typing import Any, Type

import pytest

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, Document


class TestPropertyType(JsonProperty):
    FIRST_NAME = auto()
    SECOND_NAME = auto()


@dataclass_with_properties
class TestDataModelType:
    first_property: str
    second_property: int
    third_property: int

    def __init__(self, first_property: str, second_property: int, third_property: int):
        check_types_and_set_values(self, locals())


class TestConverter(TypedConverter):
    def json_property_name(self, test_property: TestPropertyType) -> str:
        if test_property == TestPropertyType.FIRST_NAME:
            return "jsonFirstName"
        else:
            return "jsonSecondName"

    def _get_property_value(
        self, instance: TestDataModelType, test_property: TestPropertyType, _document: Document = None
    ) -> Any:
        if test_property == TestPropertyType.FIRST_NAME:
            return instance.first_property
        elif test_property == TestPropertyType.SECOND_NAME:
            return instance.second_property + instance.third_property

    def get_json_type(self) -> Type[JsonProperty]:
        return TestPropertyType

    def get_data_model_type(self) -> Type:
        return TestDataModelType


def test_conversion():
    converter = TestConverter()
    test_instance = TestDataModelType("firstPropertyValue", 1, 2)

    converted_dict = converter.convert(test_instance)

    assert converted_dict == {"jsonFirstName": "firstPropertyValue", "jsonSecondName": 3}


def test_wrong_type():
    converter = TestConverter()
    checksum = Checksum(ChecksumAlgorithm.SHA1, "123")

    with pytest.raises(TypeError) as error:
        converter.convert(checksum)

    error_message: str = error.value.args[0]
    assert TestConverter.__name__ in error_message
    assert TestDataModelType.__name__ in error_message
    assert Checksum.__name__ in error_message



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from .cvss_v2_vuln_assessment_relationship import CvssV2VulnAssessmentRelationship
from .cvss_v3_vuln_assessment_relationship import CvssV3VulnAssessmentRelationship
from .epss_vuln_assessment_relationship import EpssVulnAssessmentRelationship
from .exploit_catalog_vuln_assessment_relationship import ExploitCatalogVulnAssessmentRelationship, ExploitCatalogType
from .ssvc_vuln_assessment_relationship import SsvcVulnAssessmentRelationship, SsvcDecisionType
from .vex_affected_vuln_assessment_relationship import VexAffectedVulnAssessmentRelationship
from .vex_fixed_vuln_assessment_relationship import VexFixedVulnAssessmentRelationship
from .vex_not_affected_vuln_assessment_relationship import (
    VexNotAffectedVulnAssessmentRelationship,
    VexJustificationType,
)
from .vex_under_investigation_vuln_assessment_relationship import VexUnderInvestigationVulnAssessmentRelationship
from .vex_vuln_assessment_relationship import VexVulnAssessmentRelationship
from .vuln_assessment_relationship import VulnAssessmentRelationship
from .vulnerability import Vulnerability



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx.model import Relationship
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_text_value, write_value


def write_relationship(relationship: Relationship, text_output: TextIO):
    write_value(
        "Relationship",
        " ".join(
            [
                relationship.spdx_element_id,
                relationship.relationship_type.name,
                str(relationship.related_spdx_element_id),
            ]
        ),
        text_output,
    )
    write_text_value("RelationshipComment", relationship.comment, text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto


class ProfileIdentifierType(Enum):
    CORE = auto()
    SOFTWARE = auto()
    LICENSING = auto()
    SECURITY = auto()
    BUILD = auto()
    AI = auto()
    DATASET = auto()
    USAGE = auto()
    EXTENSION = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os
from typing import List, Optional

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import CreationInfo, Document, Relationship, RelationshipType
from spdx_tools.spdx.parser.parse_anything import parse_file
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import creation_info_fixture, document_fixture, file_fixture, package_fixture, snippet_fixture


def test_valid_document():
    document = document_fixture()
    validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)

    assert validation_messages == []


def test_spdx_lite_validation():
    document = parse_file(os.path.join(os.path.dirname(__file__), "../data/SPDXLite.spdx"))

    assert validate_full_spdx_document(document) == []


@pytest.mark.parametrize(
    "creation_info, version_input, expected_message",
    [
        (creation_info_fixture(spdx_version="SPDX-2.3"), "SPDX-2.3", None),
        (creation_info_fixture(spdx_version="SPDX-2.3"), None, None),
        (
            creation_info_fixture(spdx_version="SPDX-2.3"),
            "SPDX-2.2",
            "provided SPDX version SPDX-2.2 does not match the document's SPDX version SPDX-2.3",
        ),
        (
            creation_info_fixture(spdx_version="SPDX-2.3"),
            "SPDX2.3",
            "provided SPDX version SPDX2.3 does not match the document's SPDX version SPDX-2.3",
        ),
        (
            creation_info_fixture(spdx_version="SPDX2.3"),
            "SPDX-2.3",
            'only SPDX versions "SPDX-2.2" and "SPDX-2.3" are supported, but the document\'s spdx_version is: SPDX2.3',
        ),
        (
            creation_info_fixture(spdx_version="SPDX2.3"),
            None,
            'only SPDX versions "SPDX-2.2" and "SPDX-2.3" are supported, but the document\'s spdx_version is: SPDX2.3',
        ),
        (
            creation_info_fixture(spdx_version="SPDX2.3"),
            "SPDX2.3",
            'only SPDX versions "SPDX-2.2" and "SPDX-2.3" are supported, but the document\'s spdx_version is: SPDX2.3',
        ),
        (
            creation_info_fixture(spdx_version="SPDX-2.1"),
            "SPDX-2.1",
            'only SPDX versions "SPDX-2.2" and "SPDX-2.3" are supported, but the document\'s '
            "spdx_version is: SPDX-2.1",
        ),
    ],
)
def test_spdx_version_handling(creation_info: CreationInfo, version_input: str, expected_message: Optional[str]):
    document: Document = document_fixture(creation_info=creation_info)
    validation_messages: List[ValidationMessage] = validate_full_spdx_document(document, version_input)

    context = ValidationContext(spdx_id=creation_info.spdx_id, element_type=SpdxElementType.DOCUMENT)
    expected: List[ValidationMessage] = []

    if expected_message:
        expected.append(ValidationMessage(expected_message, context))
        expected.append(
            ValidationMessage(
                "There are issues concerning the SPDX version of the document. "
                "As subsequent validation relies on the correct version, "
                "the validation process has been cancelled.",
                context,
            )
        )

    assert validation_messages == expected


@pytest.mark.parametrize(
    "relationships",
    [
        [Relationship(DOCUMENT_SPDX_ID, RelationshipType.DESCRIBES, "SPDXRef-File")],
        [Relationship("SPDXRef-File", RelationshipType.DESCRIBED_BY, DOCUMENT_SPDX_ID)],
    ],
)
def test_document_describes_at_least_one_element(relationships):
    document = document_fixture(relationships=relationships)
    validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)

    assert validation_messages == []


def test_document_does_not_describe_an_element_with_only_one_package():
    document = document_fixture(
        packages=[package_fixture()],
        files=[],
        snippets=[],
        relationships=[],
        annotations=[],
    )
    validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)

    assert validation_messages == []


def test_document_does_not_describe_an_element_with_multiple_elements():
    document = document_fixture(
        relationships=[Relationship("SPDXRef-Package", RelationshipType.DESCRIBES, "SPDXRef-File")]
    )
    validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)

    assert validation_messages == [
        ValidationMessage(
            f'there must be at least one relationship "{DOCUMENT_SPDX_ID} DESCRIBES ..." or "... DESCRIBED_BY '
            f'{DOCUMENT_SPDX_ID}" when there is not only a single package present',
            ValidationContext(spdx_id=DOCUMENT_SPDX_ID, element_type=SpdxElementType.DOCUMENT),
        )
    ]


def test_duplicated_spdx_ids():
    document = document_fixture(
        files=[
            file_fixture(spdx_id="SPDXRef-File"),
            file_fixture(spdx_id="SPDXRef-2"),
            file_fixture(spdx_id="SPDXRef-3"),
        ],
        packages=[package_fixture(spdx_id="SPDXRef-2"), package_fixture(spdx_id=DOCUMENT_SPDX_ID)],
        snippets=[snippet_fixture(spdx_id="SPDXRef-2"), snippet_fixture(spdx_id="SPDXRef-3")],
    )

    context = ValidationContext(spdx_id=document.creation_info.spdx_id, element_type=SpdxElementType.DOCUMENT)

    validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)

    assert validation_messages == [
        ValidationMessage(
            "every spdx_id must be unique within the document, but found the following duplicates: ['SPDXRef-2', "
            f"'SPDXRef-3', '{DOCUMENT_SPDX_ID}']",
            context,
        )
    ]



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Union
from unittest import mock
from unittest.mock import MagicMock, NonCallableMagicMock

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.file_converter import FileConverter
from spdx_tools.spdx.jsonschema.file_properties import FileProperty
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Annotation,
    AnnotationType,
    Checksum,
    ChecksumAlgorithm,
    Document,
    File,
    FileType,
    SpdxNoAssertion,
    SpdxNone,
)
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING
from spdx_tools.spdx.model.spdx_none import SPDX_NONE_STRING
from tests.spdx.fixtures import annotation_fixture, creation_info_fixture, document_fixture, file_fixture
from tests.spdx.mock_utils import assert_mock_method_called_with_arguments


@pytest.fixture
@mock.patch("spdx_tools.spdx.jsonschema.checksum_converter.ChecksumConverter", autospec=True)
@mock.patch("spdx_tools.spdx.jsonschema.annotation_converter.AnnotationConverter", autospec=True)
def converter(annotation_converter_mock: MagicMock, checksum_converter_mock: MagicMock) -> FileConverter:
    converter = FileConverter()
    converter.checksum_converter = checksum_converter_mock()
    converter.annotation_converter = annotation_converter_mock()
    return converter


@pytest.mark.parametrize(
    "file_property,expected",
    [
        (FileProperty.SPDX_ID, "SPDXID"),
        (FileProperty.ANNOTATIONS, "annotations"),
        (FileProperty.ARTIFACT_OFS, "artifactOfs"),
        (FileProperty.ATTRIBUTION_TEXTS, "attributionTexts"),
        (FileProperty.CHECKSUMS, "checksums"),
        (FileProperty.COMMENT, "comment"),
        (FileProperty.COPYRIGHT_TEXT, "copyrightText"),
        (FileProperty.FILE_CONTRIBUTORS, "fileContributors"),
        (FileProperty.FILE_DEPENDENCIES, "fileDependencies"),
        (FileProperty.FILE_NAME, "fileName"),
        (FileProperty.FILE_TYPES, "fileTypes"),
        (FileProperty.LICENSE_COMMENTS, "licenseComments"),
        (FileProperty.LICENSE_CONCLUDED, "licenseConcluded"),
        (FileProperty.LICENSE_INFO_IN_FILES, "licenseInfoInFiles"),
        (FileProperty.NOTICE_TEXT, "noticeText"),
    ],
)
def test_json_property_names(converter: FileConverter, file_property: FileProperty, expected: str):
    assert converter.json_property_name(file_property) == expected


def test_json_type(converter: FileConverter):
    assert converter.get_json_type() == FileProperty


def test_data_model_type(converter: FileConverter):
    assert converter.get_data_model_type() == File


def test_successful_conversion(converter: FileConverter):
    converter.checksum_converter.convert.return_value = "mock_converted_checksum"
    converter.annotation_converter.convert.return_value = "mock_converted_annotation"
    file = File(
        name="name",
        spdx_id="spdxId",
        checksums=[Checksum(ChecksumAlgorithm.SHA224, "sha224"), Checksum(ChecksumAlgorithm.MD2, "md2")],
        file_types=[FileType.SPDX, FileType.OTHER],
        license_concluded=spdx_licensing.parse("MIT and GPL-2.0"),
        license_info_in_file=[spdx_licensing.parse("MIT"), spdx_licensing.parse("GPL-2.0"), SpdxNoAssertion()],
        license_comment="licenseComment",
        copyright_text="copyrightText",
        comment="comment",
        notice="notice",
        contributors=["contributor1", "contributor2"],
        attribution_texts=["attributionText1", "attributionText2"],
    )

    annotations = [
        Annotation(
            file.spdx_id,
            AnnotationType.REVIEW,
            Actor(ActorType.PERSON, "annotatorName"),
            datetime(2022, 12, 5),
            "review comment",
        )
    ]
    document = Document(creation_info_fixture(), files=[file], annotations=annotations)

    converted_dict = converter.convert(file, document)

    assert converted_dict == {
        converter.json_property_name(FileProperty.SPDX_ID): "spdxId",
        converter.json_property_name(FileProperty.ANNOTATIONS): ["mock_converted_annotation"],
        converter.json_property_name(FileProperty.ATTRIBUTION_TEXTS): ["attributionText1", "attributionText2"],
        converter.json_property_name(FileProperty.CHECKSUMS): ["mock_converted_checksum", "mock_converted_checksum"],
        converter.json_property_name(FileProperty.COMMENT): "comment",
        converter.json_property_name(FileProperty.COPYRIGHT_TEXT): "copyrightText",
        converter.json_property_name(FileProperty.FILE_CONTRIBUTORS): ["contributor1", "contributor2"],
        converter.json_property_name(FileProperty.FILE_NAME): "name",
        converter.json_property_name(FileProperty.FILE_TYPES): ["SPDX", "OTHER"],
        converter.json_property_name(FileProperty.LICENSE_COMMENTS): "licenseComment",
        converter.json_property_name(FileProperty.LICENSE_CONCLUDED): "MIT AND GPL-2.0-only",
        converter.json_property_name(FileProperty.LICENSE_INFO_IN_FILES): ["MIT", "GPL-2.0-only", "NOASSERTION"],
        converter.json_property_name(FileProperty.NOTICE_TEXT): "notice",
    }


def test_null_values(converter: FileConverter):
    file = file_fixture(
        copyright_text=None,
        license_concluded=None,
        license_comment=None,
        comment=None,
        notice=None,
        attribution_texts=[],
        checksums=[],
        contributors=[],
        file_types=[],
        license_info_in_file=[],
    )
    document = Document(creation_info_fixture(), files=[file])

    converted_dict = converter.convert(file, document)

    assert converter.json_property_name(FileProperty.COPYRIGHT_TEXT) not in converted_dict
    assert converter.json_property_name(FileProperty.LICENSE_CONCLUDED) not in converted_dict
    assert converter.json_property_name(FileProperty.LICENSE_COMMENTS) not in converted_dict
    assert converter.json_property_name(FileProperty.COMMENT) not in converted_dict
    assert converter.json_property_name(FileProperty.NOTICE_TEXT) not in converted_dict
    assert converter.json_property_name(FileProperty.ANNOTATIONS) not in converted_dict
    assert converter.json_property_name(FileProperty.ATTRIBUTION_TEXTS) not in converted_dict
    assert converter.json_property_name(FileProperty.CHECKSUMS) not in converted_dict
    assert converter.json_property_name(FileProperty.FILE_CONTRIBUTORS) not in converted_dict
    assert converter.json_property_name(FileProperty.FILE_TYPES) not in converted_dict
    assert converter.json_property_name(FileProperty.LICENSE_INFO_IN_FILES) not in converted_dict


def test_spdx_no_assertion(converter: FileConverter):
    file = file_fixture(
        license_concluded=SpdxNoAssertion(), license_info_in_file=[SpdxNoAssertion()], copyright_text=SpdxNoAssertion()
    )
    document = Document(creation_info_fixture(), files=[file])

    converted_dict = converter.convert(file, document)

    assert converted_dict[converter.json_property_name(FileProperty.COPYRIGHT_TEXT)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(FileProperty.LICENSE_CONCLUDED)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(FileProperty.LICENSE_INFO_IN_FILES)] == [
        SPDX_NO_ASSERTION_STRING
    ]


def test_spdx_none(converter: FileConverter):
    file = file_fixture(license_concluded=SpdxNone(), license_info_in_file=[SpdxNone()], copyright_text=SpdxNone())
    document = Document(creation_info_fixture(), files=[file])

    converted_dict = converter.convert(file, document)

    assert converted_dict[converter.json_property_name(FileProperty.COPYRIGHT_TEXT)] == SPDX_NONE_STRING
    assert converted_dict[converter.json_property_name(FileProperty.LICENSE_CONCLUDED)] == SPDX_NONE_STRING
    assert converted_dict[converter.json_property_name(FileProperty.LICENSE_INFO_IN_FILES)] == [SPDX_NONE_STRING]


def test_file_annotations(converter: FileConverter):
    file = file_fixture(spdx_id="fileId")
    document = document_fixture(files=[file])
    first_file_annotation = annotation_fixture(spdx_id=file.spdx_id)
    second_file_annotation = annotation_fixture(spdx_id=file.spdx_id)
    document_annotation = annotation_fixture(spdx_id=document.creation_info.spdx_id)
    package_annotation = annotation_fixture(spdx_id=document.packages[0].spdx_id)
    snippet_annotation = annotation_fixture(spdx_id=document.snippets[0].spdx_id)
    other_annotation = annotation_fixture(spdx_id="otherId")
    annotations = [
        first_file_annotation,
        second_file_annotation,
        document_annotation,
        package_annotation,
        snippet_annotation,
        other_annotation,
    ]
    document.annotations = annotations

    # Weird type hint to make warnings about unresolved references from the mock class disappear
    annotation_converter: Union[AnnotationConverter, NonCallableMagicMock] = converter.annotation_converter
    annotation_converter.convert.return_value = "mock_converted_annotation"

    converted_dict = converter.convert(file, document)

    assert_mock_method_called_with_arguments(
        annotation_converter, "convert", first_file_annotation, second_file_annotation
    )
    converted_file_annotations = converted_dict.get(converter.json_property_name(FileProperty.ANNOTATIONS))
    assert converted_file_annotations == ["mock_converted_annotation", "mock_converted_annotation"]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest
from rdflib import RDF, Graph, Literal, URIRef

from spdx_tools.spdx.model import ChecksumAlgorithm
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.checksum_writer import add_checksum_to_graph, algorithm_to_rdf_string
from tests.spdx.fixtures import checksum_fixture


def test_add_checksum_to_graph():
    graph = Graph()
    checksum = checksum_fixture()

    add_checksum_to_graph(checksum, graph, URIRef("parentNode"))

    assert (URIRef("parentNode"), SPDX_NAMESPACE.checksum, None) in graph
    assert (None, RDF.type, SPDX_NAMESPACE.Checksum) in graph
    assert (None, SPDX_NAMESPACE.algorithm, SPDX_NAMESPACE.checksumAlgorithm_sha1) in graph
    assert (None, SPDX_NAMESPACE.checksumValue, Literal(checksum.value)) in graph


@pytest.mark.parametrize(
    "algorithm,expected",
    [
        (ChecksumAlgorithm.SHA1, SPDX_NAMESPACE.checksumAlgorithm_sha1),
        (ChecksumAlgorithm.SHA224, SPDX_NAMESPACE.checksumAlgorithm_sha224),
        (ChecksumAlgorithm.SHA256, SPDX_NAMESPACE.checksumAlgorithm_sha256),
        (ChecksumAlgorithm.SHA384, SPDX_NAMESPACE.checksumAlgorithm_sha384),
        (ChecksumAlgorithm.SHA512, SPDX_NAMESPACE.checksumAlgorithm_sha512),
        (ChecksumAlgorithm.SHA3_256, SPDX_NAMESPACE.checksumAlgorithm_sha3_256),
        (ChecksumAlgorithm.SHA3_384, SPDX_NAMESPACE.checksumAlgorithm_sha3_384),
        (ChecksumAlgorithm.SHA3_512, SPDX_NAMESPACE.checksumAlgorithm_sha3_512),
        (ChecksumAlgorithm.BLAKE2B_256, SPDX_NAMESPACE.checksumAlgorithm_blake2b256),
        (ChecksumAlgorithm.BLAKE2B_384, SPDX_NAMESPACE.checksumAlgorithm_blake2b384),
        (ChecksumAlgorithm.BLAKE2B_512, SPDX_NAMESPACE.checksumAlgorithm_blake2b512),
        (ChecksumAlgorithm.BLAKE3, SPDX_NAMESPACE.checksumAlgorithm_blake3),
        (ChecksumAlgorithm.MD2, SPDX_NAMESPACE.checksumAlgorithm_md2),
        (ChecksumAlgorithm.MD4, SPDX_NAMESPACE.checksumAlgorithm_md4),
        (ChecksumAlgorithm.MD5, SPDX_NAMESPACE.checksumAlgorithm_md5),
        (ChecksumAlgorithm.MD6, SPDX_NAMESPACE.checksumAlgorithm_md6),
        (ChecksumAlgorithm.ADLER32, SPDX_NAMESPACE.checksumAlgorithm_adler32),
    ],
)
def test_algorithm_to_rdf_string(algorithm, expected):
    rdf_element = algorithm_to_rdf_string(algorithm)

    assert rdf_element == expected



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.model import Package, PackageVerificationCode
from spdx_tools.spdx.writer.tagvalue.checksum_writer import write_checksum_to_tag_value
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import (
    transform_enum_name_to_tv,
    write_actor,
    write_text_value,
    write_value,
)


def write_package(package: Package, text_output: TextIO):
    text_output.write("## Package Information\n")

    write_value("PackageName", package.name, text_output)
    write_value("SPDXID", package.spdx_id, text_output)
    write_value("PackageVersion", package.version, text_output)
    write_value("PackageFileName", package.file_name, text_output)
    write_actor("PackageSupplier", package.supplier, text_output)
    write_actor("PackageOriginator", package.originator, text_output)
    write_value("PackageDownloadLocation", package.download_location, text_output)

    write_value("FilesAnalyzed", str(package.files_analyzed).lower(), text_output)
    if package.verification_code:
        package_verification_code = get_package_verification_code_string(package.verification_code)
        write_value("PackageVerificationCode", package_verification_code, text_output)

    for package_checksum in package.checksums:
        write_value("PackageChecksum", write_checksum_to_tag_value(package_checksum), text_output)

    write_value("PackageHomePage", package.homepage, text_output)
    write_text_value("PackageSourceInfo", package.source_info, text_output)

    write_value("PackageLicenseConcluded", package.license_concluded, text_output)
    for license_info in package.license_info_from_files:
        write_value("PackageLicenseInfoFromFiles", license_info, text_output)
    write_value("PackageLicenseDeclared", package.license_declared, text_output)
    write_text_value("PackageLicenseComments", package.license_comment, text_output)
    write_text_value("PackageCopyrightText", package.copyright_text, text_output)

    write_text_value("PackageSummary", package.summary, text_output)
    write_text_value("PackageDescription", package.description, text_output)
    write_text_value("PackageComment", package.comment, text_output)

    for external_reference in package.external_references:
        external_reference_str = " ".join(
            [
                transform_enum_name_to_tv(external_reference.category.name),
                external_reference.reference_type,
                external_reference.locator,
            ]
        )
        write_value("ExternalRef", external_reference_str, text_output)
        if external_reference.comment:
            write_text_value("ExternalRefComment", external_reference.comment, text_output)

    for attribution_text in package.attribution_texts:
        write_text_value("PackageAttributionText", attribution_text, text_output)

    if package.primary_package_purpose:
        write_value(
            "PrimaryPackagePurpose", transform_enum_name_to_tv(package.primary_package_purpose.name), text_output
        )

    if package.release_date:
        write_value("ReleaseDate", datetime_to_iso_string(package.release_date), text_output)
    if package.built_date:
        write_value("BuiltDate", datetime_to_iso_string(package.built_date), text_output)
    if package.valid_until_date:
        write_value("ValidUntilDate", datetime_to_iso_string(package.valid_until_date), text_output)


def get_package_verification_code_string(verification_code: PackageVerificationCode) -> str:
    if not verification_code.excluded_files:
        return verification_code.value

    excluded_files_str = " (excludes: " + " ".join(verification_code.excluded_files) + ")"
    return verification_code.value + excluded_files_str



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from dataclasses import field
from datetime import datetime

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.spdx3.model import Element


@dataclass_with_properties
class Artifact(Element):
    originated_by: List[str] = field(default_factory=list)  # SPDXID of the Agent/Tool
    supplied_by: List[str] = field(default_factory=list)  # SPDXID of the Agent/Tool
    built_time: Optional[datetime] = None
    release_time: Optional[datetime] = None
    valid_until_time: Optional[datetime] = None
    standard: List[str] = field(default_factory=list)

    @abstractmethod
    def __init__(self):
        pass



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import TextIO

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.model import CreationInfo
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import (
    write_optional_heading,
    write_separator,
    write_text_value,
    write_value,
)


def write_creation_info(creation_info: CreationInfo, text_output: TextIO):
    write_value("SPDXVersion", creation_info.spdx_version, text_output)
    write_value("DataLicense", creation_info.data_license, text_output)
    write_value("SPDXID", creation_info.spdx_id, text_output)
    write_value("DocumentName", creation_info.name, text_output)
    write_value("DocumentNamespace", creation_info.document_namespace, text_output)
    write_text_value("DocumentComment", creation_info.document_comment, text_output)

    write_optional_heading(creation_info.external_document_refs, "\n## External Document References\n", text_output)
    for external_document_ref in creation_info.external_document_refs:
        external_document_ref_str = " ".join(
            [
                external_document_ref.document_ref_id,
                external_document_ref.document_uri,
                external_document_ref.checksum.algorithm.name + ": " + external_document_ref.checksum.value,
            ]
        )
        write_value("ExternalDocumentRef", external_document_ref_str, text_output)
    write_separator(text_output)

    text_output.write("## Creation Information\n")
    write_value("LicenseListVersion", creation_info.license_list_version, text_output)
    for creator in creation_info.creators:
        write_value("Creator", creator.to_serialized_string(), text_output)
    write_value("Created", datetime_to_iso_string(creation_info.created), text_output)
    write_text_value("CreatorComment", creation_info.creator_comment, text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model.ai import AIPackage
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.software.package_writer import write_package


def write_ai_package(ai_package: AIPackage, text_output: TextIO):
    text_output.write("## AI Package\n")
    write_package(ai_package, text_output, False)

    for property_name in AIPackage.__annotations__.keys():
        write_value(property_name, getattr(ai_package, property_name), text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from .any_license_info import AnyLicenseInfo
from .conjunctive_license_set import ConjunctiveLicenseSet
from .custom_license import CustomLicense
from .custom_license_addition import CustomLicenseAddition
from .disjunctive_license_set import DisjunctiveLicenseSet
from .license import License
from .license_addition import LicenseAddition
from .license_field import LicenseField
from .listed_license import ListedLicense
from .listed_license_exception import ListedLicenseException
from .no_assertion_license import NoAssertionLicense
from .none_license import NoneLicense
from .or_later_operator import OrLaterOperator
from .with_addition_operator import WithAdditionOperator



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from unittest import TestCase

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Checksum,
    ChecksumAlgorithm,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    PackagePurpose,
    PackageVerificationCode,
    SpdxNoAssertion,
    SpdxNone,
)
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.jsonlikedict.dict_parsing_functions import parse_list_of_elements
from spdx_tools.spdx.parser.jsonlikedict.package_parser import PackageParser


@pytest.mark.parametrize(
    "homepage, expected_homepage, download_location, expected_download_location, "
    "copyright_text, expected_copyright_text, originator, expected_originator, supplier, expected_supplier",
    [
        (
            "http://ftp.gnu.org/gnu/glibc",
            "http://ftp.gnu.org/gnu/glibc",
            "NOASSERTION",
            SpdxNoAssertion(),
            "NONE",
            SpdxNone(),
            "Organization: ExampleCodeInspect (contact@example.com)",
            Actor(ActorType.ORGANIZATION, "ExampleCodeInspect", "contact@example.com"),
            "NOASSERTION",
            SpdxNoAssertion(),
        ),
        (
            "NOASSERTION",
            SpdxNoAssertion(),
            "NONE",
            SpdxNone(),
            "Copyright 2008-2010 John Smith",
            "Copyright 2008-2010 John Smith",
            None,
            None,
            None,
            None,
        ),
        (
            "NONE",
            SpdxNone(),
            "http://ftp.gnu.org/gnu/glibc/glibc-ports-2.15.tar.gz",
            "http://ftp.gnu.org/gnu/glibc/glibc-ports-2.15.tar.gz",
            "NOASSERTION",
            SpdxNoAssertion(),
            "NOASSERTION",
            SpdxNoAssertion(),
            "Person: Jane Doe (jane.doe@example.com)",
            Actor(ActorType.PERSON, "Jane Doe", "jane.doe@example.com"),
        ),
    ],
)
def test_parse_package(
    homepage,
    expected_homepage,
    download_location,
    expected_download_location,
    copyright_text,
    expected_copyright_text,
    originator,
    expected_originator,
    supplier,
    expected_supplier,
):
    package_parser = PackageParser()

    package_dict = {
        "SPDXID": "SPDXRef-Package",
        "attributionTexts": [
            "The GNU C Library is free software.  See the file COPYING.LIB for copying conditions, and LICENSES for "
            "notices about a few contributions that require these additional notices to be distributed.  License "
            "copyright years may be listed using range notation, e.g., 1996-2015, indicating that every year in the "
            "range, inclusive, is a copyrightable year that would otherwise be listed individually."
        ],
        "builtDate": "2011-01-29T18:30:22Z",
        "checksums": [
            {"algorithm": "MD5", "checksumValue": "624c1abb3664f4b35547e7c73864ad24"},
            {"algorithm": "SHA1", "checksumValue": "85ed0817af83a24ad8da68c2b5094de69833983c"},
            {
                "algorithm": "SHA256",
                "checksumValue": "11b6d3ee554eedf79299905a98f9b9a04e498210b59f15094c916c91d150efcd",
            },
            {
                "algorithm": "BLAKE2b-384",
                "checksumValue": "aaabd89c926ab525c242e6621f2f5fa73aa4afe3d9e24aed727faaadd6af38b620bdb623dd2b4788b1c8"
                "086984af8706",
            },
        ],
        "comment": "This is a comment.",
        "copyrightText": copyright_text,
        "description": "The GNU C Library defines functions that are specified by the ISO C standard, as well as "
        "additional features specific to POSIX and other derivatives of the Unix operating system, and "
        "extensions specific to GNU systems.",
        "downloadLocation": download_location,
        "externalRefs": [
            {
                "referenceCategory": "SECURITY",
                "referenceLocator": "cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:*",
                "referenceType": "cpe23Type",
            },
            {
                "comment": "This is the external ref for Acme",
                "referenceCategory": "OTHER",
                "referenceLocator": "acmecorp/acmenator/4.1.3-alpha",
                "referenceType": "http://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301"
                "#LocationRef-acmeforge",
            },
        ],
        "filesAnalyzed": True,
        "homepage": homepage,
        "licenseComments": "The license for this project changed with the release of version x.y.  The version of the "
        "project included here post-dates the license change.",
        "licenseConcluded": "(LGPL-2.0-only OR LicenseRef-3)",
        "licenseDeclared": "(LGPL-2.0-only AND LicenseRef-3)",
        "licenseInfoFromFiles": ["GPL-2.0-only", "LicenseRef-2", "LicenseRef-1", "NOASSERTION"],
        "name": "glibc",
        "originator": originator,
        "packageFileName": "glibc-2.11.1.tar.gz",
        "packageVerificationCode": {
            "packageVerificationCodeExcludedFiles": ["./package.spdx"],
            "packageVerificationCodeValue": "d6a770ba38583ed4bb4525bd96e50461655d2758",
        },
        "primaryPackagePurpose": "SOURCE",
        "releaseDate": "2012-01-29T18:30:22Z",
        "sourceInfo": "uses glibc-2_11-branch from git://sourceware.org/git/glibc.git.",
        "summary": "GNU C library.",
        "supplier": supplier,
        "validUntilDate": "2014-01-29T18:30:22Z",
        "versionInfo": "2.11.1",
    }

    package = package_parser.parse_package(package_dict)

    assert package.spdx_id == "SPDXRef-Package"
    assert package.name == "glibc"
    assert package.download_location == expected_download_location
    assert package.version == "2.11.1"
    assert package.file_name == "glibc-2.11.1.tar.gz"
    assert package.supplier == expected_supplier
    assert package.originator == expected_originator
    assert package.files_analyzed is True
    assert package.verification_code == PackageVerificationCode(
        value="d6a770ba38583ed4bb4525bd96e50461655d2758", excluded_files=["./package.spdx"]
    )
    assert len(package.checksums) == 4
    TestCase().assertCountEqual(
        package.checksums,
        [
            Checksum(ChecksumAlgorithm.MD5, "624c1abb3664f4b35547e7c73864ad24"),
            Checksum(ChecksumAlgorithm.SHA1, "85ed0817af83a24ad8da68c2b5094de69833983c"),
            Checksum(ChecksumAlgorithm.SHA256, "11b6d3ee554eedf79299905a98f9b9a04e498210b59f15094c916c91d150efcd"),
            Checksum(
                ChecksumAlgorithm.BLAKE2B_384,
                "aaabd89c926ab525c242e6621f2f5fa73aa4afe3d9e24aed727faaadd6af38b620bdb623dd2b4788b1c8086984af8706",
            ),
        ],
    )
    assert package.homepage == expected_homepage
    assert package.source_info == "uses glibc-2_11-branch from git://sourceware.org/git/glibc.git."
    assert package.license_concluded == spdx_licensing.parse("(LGPL-2.0-only OR LicenseRef-3)")
    TestCase().assertCountEqual(
        package.license_info_from_files,
        [
            spdx_licensing.parse("GPL-2.0-only"),
            spdx_licensing.parse("LicenseRef-2"),
            spdx_licensing.parse("LicenseRef-1"),
            SpdxNoAssertion(),
        ],
    )
    assert package.license_declared == spdx_licensing.parse("(LGPL-2.0-only AND LicenseRef-3)")
    assert (
        package.license_comment
        == "The license for this project changed with the release of version x.y.  The version of the project included"
        " here post-dates the license change."
    )
    assert package.copyright_text == expected_copyright_text
    assert package.summary == "GNU C library."
    assert (
        package.description
        == "The GNU C Library defines functions that are specified by the ISO C standard, as well as additional "
        "features specific to POSIX and other derivatives of the Unix operating system, and extensions specific "
        "to GNU systems."
    )
    assert package.comment == "This is a comment."
    assert len(package.external_references) == 2
    TestCase().assertCountEqual(
        package.external_references,
        [
            ExternalPackageRef(
                ExternalPackageRefCategory.SECURITY,
                "cpe23Type",
                "cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:*",
            ),
            ExternalPackageRef(
                ExternalPackageRefCategory.OTHER,
                "http://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301#LocationRef-acmeforge",
                locator="acmecorp/acmenator/4.1.3-alpha",
                comment="This is the external ref for Acme",
            ),
        ],
    )
    assert package.attribution_texts == [
        "The GNU C Library is free software.  See the file COPYING.LIB for copying conditions, and LICENSES for "
        "notices about a few contributions that require these additional notices to be distributed.  License "
        "copyright years may be listed using range notation, e.g., 1996-2015, indicating that every year in the "
        "range, inclusive, is a copyrightable year that would otherwise be listed individually."
    ]
    assert package.primary_package_purpose == PackagePurpose.SOURCE
    assert package.release_date == datetime(2012, 1, 29, 18, 30, 22)
    assert package.built_date == datetime(2011, 1, 29, 18, 30, 22)
    assert package.valid_until_date == datetime(2014, 1, 29, 18, 30, 22)


@pytest.mark.parametrize(
    "incomplete_package_dict",
    [
        {"SPDXID": "SPDXRef-Package"},
        {"SPDXID": "SPDXRef-Package", "name": 5, "downloadLocation": "NONE"},
        {
            "SPDXID": "SPDXRef-Package",
            "name": "Example Package",
            "downloadLocation": "NONE",
            "checksums": [{"algorithm": "SHA", "value": "1234"}],
        },
    ],
)
def test_parse_invalid_package(incomplete_package_dict):
    package_parser = PackageParser()

    with pytest.raises(SPDXParsingError):
        package_parser.parse_package(incomplete_package_dict)


def test_parse_packages():
    package_parser = PackageParser()
    packages_list = [
        {
            "SPDXID": "SPDXRef-Package",
            "name": "Example Package",
            "downloadLocation": "NONE",
            "checksums": [{"algorithm": "SHA", "value": "1234"}],
        },
        {"SPDXID": "SPDXRef-Package", "name": 5, "downloadLocation": "NONE"},
        {"SPDXID": "SPDXRef-Package", "name": "Example Package", "downloadLocation": "NONE"},
    ]

    with pytest.raises(SPDXParsingError):
        parse_list_of_elements(packages_list, package_parser.parse_package)


def test_parse_external_ref():
    package_parser = PackageParser()
    external_ref = {"referenceType": "fix"}

    with pytest.raises(SPDXParsingError):
        package_parser.parse_external_ref(external_ref)


def test_parse_invalid_external_package_ref_category():
    package_parser = PackageParser()
    external_package_ref_category = "TEST"

    with pytest.raises(SPDXParsingError) as err:
        package_parser.parse_external_ref_category(external_package_ref_category)

    TestCase().assertCountEqual(err.value.get_messages(), ["Invalid ExternalPackageRefCategory: TEST"])



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model import NamespaceMap
from tests.spdx3.fixtures import namespace_map_fixture
from tests.spdx3.model.model_test_utils import get_property_names


def test_correct_initialization():
    namespace_map = namespace_map_fixture()

    for property_name in get_property_names(NamespaceMap):
        assert getattr(namespace_map, property_name) is not None

    assert namespace_map.prefix == "namespaceMapPrefix"
    assert namespace_map.namespace == "https://spdx.test/tools-python/namespace_map_namespace"


def test_invalid_initialization():
    with pytest.raises(TypeError) as err:
        NamespaceMap(34, ["list of namespaces"])

    assert len(err.value.args[0]) == 2
    for error in err.value.args[0]:
        assert error.startswith("SetterError NamespaceMap:")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field

from beartype.typing import List, Optional, Union

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx.model import SpdxNoAssertion


@dataclass_with_properties
class ExtractedLicensingInfo:
    license_id: Optional[str] = None
    extracted_text: Optional[str] = None
    license_name: Optional[Union[str, SpdxNoAssertion]] = None
    cross_references: List[str] = field(default_factory=list)
    comment: Optional[str] = None

    def __init__(
        self,
        license_id: Optional[str] = None,
        extracted_text: Optional[str] = None,
        license_name: Optional[Union[str, SpdxNoAssertion]] = None,
        cross_references: List[str] = None,
        comment: Optional[str] = None,
    ):
        cross_references = [] if cross_references is None else cross_references
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import CreationInfo, ExternalIdentifier, ExternalReference, IntegrityMethod
from spdx_tools.spdx3.model.licensing import LicenseField
from spdx_tools.spdx3.model.software import SoftwarePurpose
from spdx_tools.spdx3.model.software.software_artifact import SoftwareArtifact


@dataclass_with_properties
class File(SoftwareArtifact):
    content_type: Optional[str] = None  # placeholder for MediaType

    def __init__(
        self,
        spdx_id: str,
        name: str,
        creation_info: Optional[CreationInfo] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        verified_using: List[IntegrityMethod] = None,
        external_reference: List[ExternalReference] = None,
        external_identifier: List[ExternalIdentifier] = None,
        extension: Optional[str] = None,
        originated_by: List[str] = None,
        supplied_by: List[str] = None,
        built_time: Optional[datetime] = None,
        release_time: Optional[datetime] = None,
        valid_until_time: Optional[datetime] = None,
        standard: List[str] = None,
        content_identifier: Optional[str] = None,
        primary_purpose: Optional[SoftwarePurpose] = None,
        additional_purpose: List[SoftwarePurpose] = None,
        concluded_license: Optional[LicenseField] = None,
        declared_license: Optional[LicenseField] = None,
        copyright_text: Optional[str] = None,
        attribution_text: Optional[str] = None,
        content_type: Optional[str] = None,
    ):
        verified_using = [] if verified_using is None else verified_using
        external_reference = [] if external_reference is None else external_reference
        external_identifier = [] if external_identifier is None else external_identifier
        originated_by = [] if originated_by is None else originated_by
        supplied_by = [] if supplied_by is None else supplied_by
        standard = [] if standard is None else standard
        additional_purpose = [] if additional_purpose is None else additional_purpose
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from dataclasses import field
from datetime import datetime
from enum import Enum, auto

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import CreationInfo, Element, ExternalIdentifier, ExternalReference, IntegrityMethod


class RelationshipType(Enum):
    AFFECTS = auto()
    AMENDS = auto()
    ANCESTOR = auto()
    AVAILABLE_FROM = auto()
    BUILD_DEPENDENCY = auto()
    BUILD_TOOL = auto()
    COORDINATED_BY = auto()
    CONTAINS = auto()
    CONFIG_OF = auto()
    COPY = auto()
    DATA_FILE = auto()
    DEPENDENCY_MANIFEST = auto()
    DEPENDS_ON = auto()
    DESCENDANT = auto()
    DESCRIBES = auto()
    DEV_DEPENDENCY = auto()
    DEV_TOOL = auto()
    DISTRIBUTION_ARTIFACT = auto()
    DOCUMENTATION = auto()
    DOES_NOT_AFFECT = auto()
    DYNAMIC_LINK = auto()
    EXAMPLE = auto()
    EVIDENCE_FOR = auto()
    EXPANDED_FROM_ARCHIVE = auto()
    EXPLOIT_CREATED_BY = auto()
    FILE_ADDED = auto()
    FILE_DELETED = auto()
    FILE_MODIFIED = auto()
    FIXED_BY = auto()
    FIXED_IN = auto()
    FOUND_BY = auto()
    GENERATES = auto()
    HAS_ASSESSMENT_FOR = auto()
    HAS_ASSOCIATED_VULNERABILITY = auto()
    HOST_OF = auto()
    INPUT_OF = auto()
    INVOKED_BY = auto()
    METAFILE = auto()
    ON_BEHALF_OF = auto()
    OPTIONAL_COMPONENT = auto()
    OPTIONAL_DEPENDENCY = auto()
    OTHER = auto()
    OUTPUT_OF = auto()
    PACKAGES = auto()
    PATCH = auto()
    PREREQUISITE = auto()
    PROVIDED_DEPENDENCY = auto()
    PUBLISHED_BY = auto()
    REPORTED_BY = auto()
    REPUBLISHED_BY = auto()
    REQUIREMENT_FOR = auto()
    RUNTIME_DEPENDENCY = auto()
    SPECIFICATION_FOR = auto()
    STATIC_LINK = auto()
    TEST = auto()
    TEST_CASE = auto()
    TEST_DEPENDENCY = auto()
    TEST_TOOL = auto()
    TESTED_ON = auto()
    TRAINED_ON = auto()
    UNDER_INVESTIGATION_FOR = auto()
    VARIANT = auto()


class RelationshipCompleteness(Enum):
    INCOMPLETE = auto()
    COMPLETE = auto()
    NOASSERTION = auto()


@dataclass_with_properties
class Relationship(Element):
    # due to the inheritance we need to make all fields non-default in the __annotation__,
    # the __init__ method still raises an error if required fields are not set
    from_element: str = None
    to: List[str] = field(default_factory=list)
    relationship_type: RelationshipType = None
    completeness: Optional[RelationshipCompleteness] = None
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None

    def __init__(
        self,
        spdx_id: str,
        from_element: str,
        relationship_type: RelationshipType,
        to: List[str] = None,
        creation_info: Optional[CreationInfo] = None,
        name: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        verified_using: List[IntegrityMethod] = None,
        external_reference: List[ExternalReference] = None,
        external_identifier: List[ExternalIdentifier] = None,
        extension: Optional[str] = None,
        completeness: Optional[RelationshipCompleteness] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ):
        to = [] if to is None else to
        verified_using = [] if verified_using is None else verified_using
        external_reference = [] if external_reference is None else external_reference
        external_identifier = [] if external_identifier is None else external_identifier
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model.licensing import ListedLicense, ListedLicenseException, WithAdditionOperator
from tests.spdx3.fixtures import fixture_factory


def test_valid_initialization():
    lic = fixture_factory(ListedLicense)
    lic_addition = fixture_factory(ListedLicenseException)
    with_addition_operator = WithAdditionOperator(lic, lic_addition)

    assert with_addition_operator.subject_license == lic


def test_invalid_initialization():
    with pytest.raises(TypeError) as err:
        WithAdditionOperator("License", "LicenseAddition")

    assert len(err.value.args[0]) == 2
    for arg in err.value.args[0]:
        assert arg.startswith("SetterError WithAdditionOperator:")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, RDFS, Graph, Literal, URIRef

from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.extracted_licensing_info_writer import add_extracted_licensing_info_to_graph
from tests.spdx.fixtures import extracted_licensing_info_fixture


def test_add_extracted_licensing_info_to_graph():
    graph = Graph()
    extracted_licensing_info = extracted_licensing_info_fixture()

    add_extracted_licensing_info_to_graph(extracted_licensing_info, graph, URIRef("docNode"), "docNamespace")

    assert (URIRef("docNode"), SPDX_NAMESPACE.hasExtractedLicensingInfo, None) in graph
    assert (URIRef("docNamespace#LicenseRef-1"), RDF.type, SPDX_NAMESPACE.ExtractedLicensingInfo) in graph
    assert (None, SPDX_NAMESPACE.licenseId, Literal(extracted_licensing_info.license_id)) in graph
    assert (None, SPDX_NAMESPACE.extractedText, Literal(extracted_licensing_info.extracted_text)) in graph
    assert (None, RDFS.seeAlso, Literal(extracted_licensing_info.cross_references[0])) in graph
    assert (None, SPDX_NAMESPACE.name, Literal(extracted_licensing_info.license_name)) in graph
    assert (None, RDFS.comment, Literal(extracted_licensing_info.comment)) in graph



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List
from unittest import TestCase

import pytest

from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm
from spdx_tools.spdx.validation.file_validator import validate_file, validate_file_within_document
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import document_fixture, file_fixture


def test_valid_file():
    file = file_fixture()
    validation_messages: List[ValidationMessage] = validate_file_within_document(file, "SPDX-2.3", document_fixture())

    assert validation_messages == []


@pytest.mark.parametrize(
    "file_input, spdx_id, expected_message",
    [
        (
            file_fixture(name="/invalid/file/name"),
            file_fixture().spdx_id,
            'file name must not be an absolute path starting with "/", but is: /invalid/file/name',
        ),
        (
            file_fixture(checksums=[Checksum(ChecksumAlgorithm.MD2, "d4c41ce30a517d6ce9d79c8c17bb4b66")]),
            file_fixture().spdx_id,
            "checksums must contain a SHA1 algorithm checksum, but only contains: [<ChecksumAlgorithm.MD2: 13>]",
        ),
    ],
)
def test_invalid_file(file_input, spdx_id, expected_message):
    validation_messages: List[ValidationMessage] = validate_file_within_document(
        file_input, "SPDX-2.3", document_fixture()
    )

    expected = ValidationMessage(
        expected_message,
        ValidationContext(
            spdx_id=spdx_id,
            parent_id=document_fixture().creation_info.spdx_id,
            element_type=SpdxElementType.FILE,
            full_element=file_input,
        ),
    )

    assert validation_messages == [expected]


def test_v2_2mandatory_fields():
    file = file_fixture(license_concluded=None, license_info_in_file=[], copyright_text=None)

    assert validate_file(file, "SPDX-2.3") == []

    validation_messages: List[ValidationMessage] = validate_file(file, "SPDX-2.2")

    context = ValidationContext(spdx_id=file.spdx_id, element_type=SpdxElementType.FILE, full_element=file)
    mandatory_fields = ["license_concluded", "license_info_in_file", "copyright_text"]
    expected = [ValidationMessage(f"{field} is mandatory in SPDX-2.2", context) for field in mandatory_fields]

    TestCase().assertCountEqual(validation_messages, expected)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.checksum_converter import ChecksumConverter
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.file_properties import FileProperty
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.optional_utils import apply_if_present
from spdx_tools.spdx.model import Document, File


class FileConverter(TypedConverter[File]):
    annotation_converter: AnnotationConverter
    checksum_converter: ChecksumConverter

    def __init__(self):
        self.annotation_converter = AnnotationConverter()
        self.checksum_converter = ChecksumConverter()

    def json_property_name(self, file_property: FileProperty) -> str:
        if file_property == FileProperty.SPDX_ID:
            return "SPDXID"
        return super().json_property_name(file_property)

    def _get_property_value(self, file: Any, file_property: FileProperty, document: Document = None) -> Any:
        if file_property == FileProperty.SPDX_ID:
            return file.spdx_id
        elif file_property == FileProperty.ANNOTATIONS:
            file_annotations = filter(lambda annotation: annotation.spdx_id == file.spdx_id, document.annotations)
            return [self.annotation_converter.convert(annotation) for annotation in file_annotations] or None
        elif file_property == FileProperty.ARTIFACT_OFS:
            # Deprecated property, automatically converted during parsing
            pass
        elif file_property == FileProperty.ATTRIBUTION_TEXTS:
            return file.attribution_texts or None
        elif file_property == FileProperty.CHECKSUMS:
            return [self.checksum_converter.convert(checksum) for checksum in file.checksums] or None
        elif file_property == FileProperty.COMMENT:
            return file.comment
        elif file_property == FileProperty.COPYRIGHT_TEXT:
            return apply_if_present(str, file.copyright_text)
        elif file_property == FileProperty.FILE_CONTRIBUTORS:
            return file.contributors or None
        elif file_property == FileProperty.FILE_DEPENDENCIES:
            # Deprecated property, automatically converted during parsing
            pass
        elif file_property == FileProperty.FILE_NAME:
            return file.name
        elif file_property == FileProperty.FILE_TYPES:
            return [file_type.name for file_type in file.file_types] or None
        elif file_property == FileProperty.LICENSE_COMMENTS:
            return file.license_comment
        elif file_property == FileProperty.LICENSE_CONCLUDED:
            return apply_if_present(str, file.license_concluded)
        elif file_property == FileProperty.LICENSE_INFO_IN_FILES:
            return [str(license_expression) for license_expression in file.license_info_in_file] or None
        elif file_property == FileProperty.NOTICE_TEXT:
            return file.notice

    def get_json_type(self) -> Type[JsonProperty]:
        return FileProperty

    def get_data_model_type(self) -> Type[File]:
        return File

    def requires_full_document(self) -> bool:
        return True



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from license_expression import get_spdx_licensing

# this getter takes quite long so we only call it once in this singleton module
spdx_licensing = get_spdx_licensing()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from copy import deepcopy

from beartype.typing import Any, Dict, List, Union

from spdx_tools.spdx.model import Document, File, Package, Snippet


def get_contained_spdx_element_ids(document: Document) -> List[str]:
    element_ids = [file.spdx_id for file in document.files]
    element_ids.extend([package.spdx_id for package in document.packages])
    element_ids.extend([snippet.spdx_id for snippet in document.snippets])
    return element_ids


def get_element_from_spdx_id(document: Document, spdx_id: str) -> Union[Package, File, Snippet, None]:
    contained_spdx_elements: Dict[str, Union[Package, File, Snippet]] = get_contained_spdx_elements(document)
    if spdx_id not in contained_spdx_elements:
        return None
    return contained_spdx_elements[spdx_id]


def get_contained_spdx_elements(document: Document) -> Dict[str, Union[Package, File, Snippet]]:
    contained_spdx_elements = {package.spdx_id: package for package in document.packages}
    contained_spdx_elements.update({file.spdx_id: file for file in document.files})
    contained_spdx_elements.update({snippet.spdx_id: snippet for snippet in document.snippets})

    return contained_spdx_elements


def create_document_without_duplicates(document: Document) -> Document:
    document_without_duplicates = deepcopy(document)
    for elements in [
        [document_without_duplicates.creation_info],
        document_without_duplicates.files,
        document_without_duplicates.packages,
        document_without_duplicates.snippets,
        document_without_duplicates.extracted_licensing_info,
    ]:
        for element in elements:
            for key, value in element.__dict__.items():
                if isinstance(value, list):
                    value_without_duplicates = create_list_without_duplicates(value)
                    setattr(element, key, value_without_duplicates)

    return document_without_duplicates


def create_list_without_duplicates(list_with_potential_duplicates: List[Any]) -> List[Any]:
    list_without_duplicates = []
    for element in list_with_potential_duplicates:
        if element not in list_without_duplicates:
            list_without_duplicates.append(element)

    return list_without_duplicates



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, Graph, URIRef

from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.external_document_ref_writer import add_external_document_ref_to_graph
from tests.spdx.fixtures import external_document_ref_fixture


def test_add_external_document_ref_to_graph():
    graph = Graph()
    external_document_ref = external_document_ref_fixture()

    add_external_document_ref_to_graph(external_document_ref, graph, URIRef("docNode"), "docNamespace")

    assert (
        URIRef("docNode"),
        SPDX_NAMESPACE.externalDocumentRef,
        URIRef("docNamespace#DocumentRef-external"),
    ) in graph
    assert (None, RDF.type, SPDX_NAMESPACE.ExternalDocumentRef) in graph
    assert (None, SPDX_NAMESPACE.checksum, None) in graph
    assert (None, RDF.type, SPDX_NAMESPACE.Checksum) in graph
    assert (None, SPDX_NAMESPACE.spdxDocument, URIRef(external_document_ref.document_uri)) in graph



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List, Union
from boolean import Expression
from license_expression import AND, OR, ExpressionInfo, LicenseExpression, LicenseSymbol, LicenseWithExceptionSymbol
from rdflib import RDF, BNode, Graph, URIRef
from rdflib.term import Literal, Node

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, SPDX_NAMESPACE


def add_license_expression_or_none_or_no_assertion(
    value: Union[
        List[Union[LicenseExpression, SpdxNoAssertion, SpdxNone]], LicenseExpression, SpdxNoAssertion, SpdxNone
    ],
    graph: Graph,
    parent: Node,
    predicate: Node,
    doc_namespace: str,
):
    if isinstance(value, SpdxNoAssertion):
        graph.add((parent, predicate, SPDX_NAMESPACE.noassertion))
        return
    if isinstance(value, SpdxNone):
        graph.add((parent, predicate, SPDX_NAMESPACE.none))
        return
    if isinstance(value, list):
        for license_expression in value:
            add_license_expression_or_none_or_no_assertion(license_expression, graph, parent, predicate, doc_namespace)
    if isinstance(value, LicenseExpression):
        add_license_expression_to_graph(value, graph, parent, predicate, doc_namespace)


def add_license_expression_to_graph(
    license_expression: Expression, graph: Graph, parent: Node, predicate: Node, doc_namespace: str
):
    if isinstance(license_expression, AND):
        member_node = BNode()
        graph.add((member_node, RDF.type, SPDX_NAMESPACE.ConjunctiveLicenseSet))
        graph.add((parent, predicate, member_node))
        for arg in license_expression.args:
            add_license_expression_to_graph(arg, graph, member_node, SPDX_NAMESPACE.member, doc_namespace)
    if isinstance(license_expression, OR):
        member_node = BNode()
        graph.add((member_node, RDF.type, SPDX_NAMESPACE.DisjunctiveLicenseSet))
        graph.add((parent, predicate, member_node))
        for arg in license_expression.args:
            add_license_expression_to_graph(arg, graph, member_node, SPDX_NAMESPACE.member, doc_namespace)
    if isinstance(license_expression, LicenseWithExceptionSymbol):
        member_node = BNode()
        graph.add((member_node, RDF.type, SPDX_NAMESPACE.WithExceptionOperator))
        graph.add((parent, predicate, member_node))

        add_license_expression_to_graph(
            license_expression.license_symbol, graph, member_node, SPDX_NAMESPACE.member, doc_namespace
        )
        add_license_exception_to_graph(license_expression.exception_symbol, graph, member_node)

    if isinstance(license_expression, LicenseSymbol):
        if license_or_exception_is_on_spdx_licensing_list(license_expression):
            graph.add((parent, predicate, LICENSE_NAMESPACE[str(license_expression)]))
        else:
            # assuming that the license expression is a LicenseRef to an instance of ExtractedLicensingInfo
            graph.add((parent, predicate, URIRef(f"{doc_namespace}#{license_expression}")))


def license_or_exception_is_on_spdx_licensing_list(license_symbol: LicenseSymbol) -> bool:
    symbol_info: ExpressionInfo = spdx_licensing.validate(license_symbol)
    return not symbol_info.errors


def add_license_exception_to_graph(license_exception: LicenseSymbol, graph: Graph, parent: Node):
    if license_or_exception_is_on_spdx_licensing_list(license_exception):
        exception_node = LICENSE_NAMESPACE[str(license_exception)]
        graph.add((parent, SPDX_NAMESPACE.licenseException, exception_node))
    else:
        exception_node = BNode()
        graph.add((exception_node, SPDX_NAMESPACE.licenseExceptionId, Literal(license_exception)))
        graph.add((parent, SPDX_NAMESPACE.licenseException, exception_node))

    graph.add((exception_node, RDF.type, SPDX_NAMESPACE.LicenseException))
